# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""
Defines the class used to automate the execution of the MaNGA DAP at
Utah.

Revision history
----------------

    | **?? Nov 2014**: Architecture setup by J. Brownstein
        <joelbrownstein@astro.utah.edu>
    | **18 Nov 2014**: Handover to K. Westfall (KBW) for development.
        (Attempt to) conform style to `PEP 8`_, except that lines are 99
        characters long (docstrings still limited to 72 characters).
        Also (attempt to) conform to docstrings style according to `PEP
        257`_.
    | **01 Dec 2014**: (KBW) Committed to SVN
    | **02 Dec 2014**: (KBW) Changed drpver and idlutilsver to mangaver
    | **13 Jan 2015**: (KBW) Changed it back, added platetargets
    | **05 Feb 2015**: (KBW) Change to load module based on MPL
    | **19 Mar 2015**: (KBW) Major changes to allow for more control
        over the directory structure containing the DRP data and for the
        DAP results; following changed in the drpfile and drpcomplete
        classes.
    | **27 Aug 2015**: (KBW) Sphinx documentation; prep for MPL-4;
        accommodate changes in drpcomplete; removed arg as an attribute
        because it was only used in
        :func:`mangadap.scripts.rundap._read_arg`; despite not doing
        anything before, commented out _write_module_commands for the
        time-being; removed file_root() and parameter_file() functions
        in favor of adding/using functions from
        :mod:`mangadap.config.defaults`; version number set to 1_1_0.
    | **06 Oct 2015**: (KBW) Added functionality to select the types of
        additional output.  Minor changes to allow for DR13 QA plots.
    | **22 Oct 2015**: (KBW) Added check that the selected MPL versions match
        the current environmental versions.  Includes edit to
        :class:`mangadap.survey.mangampl.MaNGAMPL` to include
        MANGACORE_VER.  Changed default number of nodes to 9.  Added
        definition of idl command for use on sciama cluster at the ICG,
        Portsmouth.
    | **04 Nov 2015**: (KBW) Added capability to turn off main DAP
        processing call so that :class:`rundap` can be used to, e.g.,
        just create the plots for existing data.
    | **17 Feb 2016**: (KBW) Added try/except block for importing
        pbs.queue; converted old drpfile to new
        :class:`mangadap.drpfits.DRPFits`, and drpcomplete to new
        :class:`mangadap.survey.drpcomplete.DRPComplete`.
    | **13 May 2016**: (KBW) Substantial changes to accommodate
        conversion of DAP from IDL to python
    | **19 May 2016**: (KBW) Added command-line options for including
        log file
    | **11 Oct 2016**: (KBW) Version checking moved to
        :class:`mangadap.survey.mangampl.MaNGAMPL`.
    | **23 Aug 2017**: (KBW) Include ppxffit QA plots.
    | **29 Aug 2017**: (KBW) Include DAPall post-processing.
    | **16 Jan 2019**: (KBW) Major changes: Only one operation mode
        (daily, all, redo removed; code always run in "redo" mode).
        Remove old "prior" options.  Setup done using DRPall by default,
        but option to use mangacore/platetargets.  'CUBE' vs. 'RSS' no
        longer an option, only analyzes CUBEs.
    | **23 Jan 2019**: (KBW) Add fit QA plots

----

.. include license and copyright
.. include:: ../include/copy.rst

----

.. include common links, assuming primary doc root is up one directory
.. include:: ../include/links.rst
"""
import sys
import time
import shutil
import os
import glob
import warnings
import argparse

try:
    import pbs.queue
except:
    warnings.warn('Could not import pbs.queue!  Any cluster submission will fail!', ImportWarning)

import numpy

from mangadap.config import defaults
from mangadap.survey.drpcomplete import DRPComplete
from mangadap.survey.mangampl import MaNGAMPL
from mangadap.util.parser import arginp_to_list
from mangadap.util.fileio import create_symlink
from mangadap.config.analysisplan import AnalysisPlanSet
from mangadap.proc.spatiallybinnedspectra import SpatiallyBinnedSpectra
from mangadap.proc.stellarcontinuummodel import StellarContinuumModel
from mangadap.proc.emissionlinemodel import EmissionLineModel


def parse_args(options=None, return_parser=False):

    # Declare the ArgumentParser object
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    # Read the optional run-mode arguments
    parser.add_argument("--overwrite",
                        help="if all selected, will run dap for all plates/ifudesigns/modes "
                        " regardless of state", action="store_true", default=False)
    parser.add_argument('-v', '--verbose', action='count',
                        help='Set verbosity level for manga_dap; can be omitted and set up '
                             'to -vv', default=0)
    parser.add_argument("--quiet", help="suppress screen output", action="store_true",
                        default=False)
    parser.add_argument("--print_version", help="print DAP version and stop",
                        action="store_true", default=False)
       
    # These arguments are used to override default behavior
    parser.add_argument("--loose", help="Only throw warnings if the versioning is "
                        "not identically as it should be for the designated MPL",
                        action="store_true", default=False)
    parser.add_argument("--mplver", type=str, help="select MPL version to analyze",
                        default=None)
    parser.add_argument("--redux_path", type=str, help="main DRP output path", default=None)
    parser.add_argument("--dapver", type=str,
                        help="optional output version, different from product version",
                        default=None)
    parser.add_argument("--analysis_path", type=str, help="main DAP output path", default=None)

    parser.add_argument("--plan_file", type=str, help="parameter file with the MaNGA DAP "
                        "execution plan to use instead of the default" , default=None)
    parser.add_argument("--platelist", type=str, help="set list of plates to reduce",
                        default=None)
    parser.add_argument("--ifudesignlist", type=str, help="set list of ifus to reduce",
                        default=None)
    parser.add_argument("--list_file", type=str,
                        help='A file with the list of plates and ifudesigns to analyze',
                        default=None)
    parser.add_argument("--combinatorics", help="force execution of all permutations of the "
                        "provided lists", action="store_true", default=False)

    parser.add_argument('--sres_ext', type=str, default=None,
                        help='Spectral resolution extension to use.  Default set by '
                             'MaNGADataCube class.')
    parser.add_argument('--sres_fill', type=str, default=None,
                        help='If present, use interpolation to fill any masked pixels in the '
                             'spectral resolution vectors. Default set by MaNGADataCube '
                             'class.')
    parser.add_argument('--covar_ext', type=str, default=None,
                        help='Use this extension to define the spatial correlation matrix.  '
                             'Default set by MaNGADataCube class.')

    parser.add_argument("--use_plttargets", action="store_true", default=False,
                        help="Use platetargets files instead of the DRPall file to "
                             "generate the DRP complete database")
    parser.add_argument("--plttargets", type=str, help="path to plateTargets file(s); "
                        "if provided will force update to drpcomplete fits file", default=None)
    parser.add_argument('--on_disk', action='store_true', default=False,
                        help='When using the DRPall file to collate the data for input to '
                             'the DAP, search for available DRP files on disk instead of '
                             'using the DRPall file content.')

    parser.add_argument("--log", help="Have the main DAP executable produce a log file",
                        action="store_true", default=False)
    parser.add_argument("--no_proc", help="Do NOT perform the main DAP processing steps",
                        action="store_true", default=False)
    parser.add_argument("--no_plots", help="Do NOT create QA plots", action="store_true",
                        default=False)
    parser.add_argument("--post", help="Create/Submit the post-processing scripts",
                        action="store_true", default=False)
    parser.add_argument("--post_plots", action="store_true", default=False,
                        help="Create/Submit the post-processing plotting scripts")

    parser.add_argument("--dapall", action="store_true", default=False,
                        help='Wait for any individual plate-ifu processes to finish and then'
                             'update the DAPall file')

    # Read arguments specific to the cluster submission behavior
    parser.add_argument("--label", type=str, help='label for cluster job', default=None)
    parser.add_argument("--nodes", type=int, help='number of nodes to use in cluster',
                        default=1)
    parser.add_argument("--cpus", type=int,
                        help='number of cpus to use per node.  Default is to use all available'
                             '; otherwise, set to minimum of provided number and number of '
                             'processors per node', default=None)
    parser.add_argument("--fast", dest='qos', type=str, help='qos state', default=None)
    parser.add_argument("--umask", type=str, help='umask bit for cluster job', default='0027')
    parser.add_argument("--walltime", type=str, help='walltime for cluster job',
                        default='240:00:00')
    parser.add_argument("--toughness", dest='hard', action='store_false', default=True,
                        help='turn off hard keyword for cluster submission')

    parser.add_argument('--create', action='store_true', default=False,
                        help='use the pbs package to create the cluster scripts')
    parser.add_argument('--submit', action='store_true', default=False,
                        help='submit the scripts to the cluster')
    parser.add_argument('--progress', action='store_true', default=False,
                        help='instead of closing the script, report the progress of the '
                             'analysis on the cluster; this is required if you want to submit '
                             'the DAPall script immediately after completing the individual '
                             'cube analysis')

    parser.add_argument("--queue", dest='queue', type=str, help='set the destination queue',
                        default=None)

    if return_parser:
        return parser

    return parser.parse_args() if options is None else parser.parse_args(options)


class rundap:
    r"""
    Class used for automated execution of the MaNGA DAP via submission
    of scripts to a cluster.

    [ More complete explanation of the class here. ]

    .. note::
    
        This is purely a survey-level utility for running the DAP at
        Utah, including submitting jobs to the cluster.  No user-level
        usage of the DAP should be considered here, apart from not
        hindering such usage of the primary python programs!

    .. todo::

        - Have :func:`_read_arg` return the boolean for :attr:`version`
          instead of keeping it as an attribute.
        - Ditch ``console`` and :func:`_read_arg` in favor of the
          more typical ``parser`` and ``main`` pairing.

    Args:
        overwrite (:obj:`bool`, optional):
            Overwrite and redo any existing analysis.
        console (:obj:`bool`, optional):
            The class has been executed from the terminal with
            command-line arguments that should be parsed.
        quiet (:obj:`bool`, optional):
            Suppress output
        print_version (:obj:`bool`, optional):
            Print the class version and return.
        strictver (:obj:`bool`, optional):
            Strictly check the version requirements of the dependencies
            defined in :mangadap.survey.mangampl.MaNGAMPL` for this
            version of the DAP.  Default is True, meaning the code will
            raise an exception if the expected version are not the same
            as the environment.
        mplver (:obj:`str`, optional):
            MPL version to analyze.  Used with
            :class:`mangadap.survey.mangampl.MaNGAMPL` to set the DAP
            dependencies.  Ideally, this controls the modules that are
            loaded before execution of the DAP.  The default version
            (selected if mplver=None) is defined by
            :class:`mangadap.survey.mangampl.MaNGAMPL`.
        redux_path (:obj:`str`, optional):
            The top-level path with the DRP files used to override the
            default defined by
            :func:`mangadap.config.defaults.drp_redux_path`.
        dapver (:obj:`str`, optional):
            The DAP version to use for the analysis, used to override
            the default defined by
            :func:`mangadap.config.defaults.dap_version`.  *Only
            used to define the file paths, not used to switch to the
            specified code version!*
        analysis_path (:obj:`str`, optional):
            The top-level path for the DAP output files, used to
            override the default defined by
            :func:`mangadap.config.defaults.dap_analysis_path`.
        plan_file (:obj:`str`, optional):
            Name of the plan file to use for *all* DRP data in this run,
            used to override the default defined by
            :func:`mangadap.par.analysisplan.AnalysisPlanSet.default`.
        platelist (:obj:`int`, :obj:`str`, optional):
            List of plates to analyze; default is to search the
            :attr:`redux_path` for any DRP files; see
            :class:`mangadap.survey.drpcomplete.DRPComplete`.
        ifudesignlist (:obj:`int`, :obj:`str`, optional):
            List of ifudesigns to analyze: default is to search the
            :attr:`redux_path` for any DRP files; see
            :class:`mangadap.survey.drpcomplete.DRPComplete`.
        combinatorics (:obj:`bool`, optional): 
            Use all unique combinations of the entered
            plate/ifudesign/mode lists to create the full list of DRP
            files to analyze.
        list_file (:obj:`str`, optional):
            File with a list of plates and ifudesigns to analyze. The
            file must have two columns, with one plate-ifu
            combination per file row.
        sres_ext (:obj:`str`, optional):
            The extension to use when constructing the
            spectral resolution vectors for the MaNGA datacubes. See
            :func:`mangadap.datacube.manga.MaNGADataCube.spectral_resolution`.
        sres_fill (:obj:`bool`, optional):
            Fill masked spectral-resolution data by simple linear
            interpolation.
        covar_ext (:obj:`str`, optional):
            Extension in the MaNGA DRP CUBE file to use as the single
            spatial correlation matrix for all wavelength channels.
        use_platetargets (:obj:`bool`, optional): 
            Flag to use the plateTargets files, instead of the DRPall
            file, as the source of the input data need by the DAP.
        platetargets (:obj:`str`, :obj:`list`, optional):
            List of platetargets files to search through to find any
            given plate-ifudesign combination.  Default is returned as
            the first element in
            :func:`mangadap.config.defaults.plate_target_files`.
        on_disk (:obj:`bool`, optional):
            When searching for available files to analyze, search the
            DRP directory path instead of using the data in the DRPall
            file.  This is implicitly true when using the plateTargets
            files to collate the data used for the DAP input instead of
            the DRPall file (i.e., when `use_platetargets=True`).
        log (:obj:`bool`, optional):
            Flag to create the DAP log files.
        dapproc (:obj:`bool`, optional):
            Flag to execute the main DAP processing.
        pltifu_plots (:obj:`bool`, optional):
            Create the QA plate-ifu specific plots.
        post_process (:obj:`bool`, optional):
            Prepare (and possibly perform) the post processing steps:
            create the DAPall file and its associated QA plots, and the
            fit QA plots for each plate.
        post_plots (:obj:`bool`, optional):
            Create the post-processing QA plots.
        report_progress (:obj:`bool`, optional):
            Report the progress of the analysis.
        vebose (:obj:`int`, optional):
            Verbosity level between 0 and 2.
        label (:obj:`str`, optional):
            Label to use in cluster queue.
        nodes (:obj:`int`, optional):
            Number of cluster nodes to use.
        qos (:obj:`str`, optional):
            Select a specific processor (e.g. sdss-fast when the manga
            user).  This option is ignored if :attr:`q` is set.
        umask (:obj:`str`, optional):
            umask to set for output.
        walltime (:obj:`str`, optional):
            Maximum wall time for cluster job.
        hard (:obj:`bool`, optional):
            Same as hard keyword in cluster submission; see
            :func:`pbs.queue.commit`.  This keyword is in the
            **opposite** sense of the "--toughness" command-line
            options; i.e. including --toughness on the command line sets
            hard=False.
        create (:obj:`bool`, optional):
            Use the pbs package to create the scripts.
        submit (:obj:`bool`, optional):
            Submit all jobs to the cluster.
        queue (:obj:`str`, optional):
            Name of the destination queue.  When submitting jobs at
            Utah, this should **not** be set (leaving it at the default
            of None).  When submitting jobs at Portsmouth, this can be
            used to select either sciama1.q, cluster.q (default), or
            sciama3.q.

    Attributes:
        overwrite (bool): **Only used with :attr:`all`**.  Overwrite and
            redo any existing analysis.
        quiet (bool): Suppress output
        print_version (bool): Print the version and then return
        strictver (bool): Strictly check the version requirements for
            this version of the dap.  If True, an exception is raised if
            the expected version are not the same as the environment.
        mpl (:class:`mangadap.survey.mangampl.MaNGAMPL`): MPL version;
            see above.
        redux_path (str): The top-level path with the DRP files.
        dapver (str): The DAP version.
        analysis_path (str): The top-level path for the DAP output
            files.
        plan_file (str): Name of the plan file to use for ALL DRP data
            in this run.  Will be None if using the default plan.
        plan (:class:`AnalysisPlanSet`): The set of analysis plans
        daptypes (:obj:`list`): This list of daptypes, one per plan
        platelist (list): List of plates to analyze.
        ifudesignlist (list): List of ifudesigns to analyze.
        combinatorics (bool): Use all unique combinations of the entered
            plate/ifudesign/mode lists to create the full list of DRP
            files to analyze.
        use_platetargets (bool): See above
        platetargets (list): List of platetargets files to search
            through to find any given plate-ifudesign combination.
        on_disk (bool): See above
        log (str): See above
        dapproc (bool): Flag to execute the main DAP processing
        pltifu_plots (bool): Create the QA plate-ifu specific plots
        post_process (:obj:`bool`): See above
        post_plots (:obj:`bool`): See above
        report_progress (:obj:`bool`): See abve
        label (str): Label to use in cluster queue.
        nodes (int): Number of cluster nodes to use.
        qos (str): Specific processor to use (Utah specific)
        umask (str): umask to set for output.
        walltime (str): Maximum wall time for cluster job.
        hard (bool): Same as hard keyword in cluster submission; see
            :func:`pbs.queue.commit`.
        create (bool): Use the pbs package to create the submission
            scripts.
        submit (bool): Submit all jobs to the cluster.
        q (str): Queue destination (Portsmouth specific)
        drpc (:class:`mangadap.survey.drpcomplete.DRPComplete`):
            Database of the available DRP files and the parameters
            necessary to write the DAP par files.

    Raises:
        FileNotFoundError:
            Raised if a plate-ifu list file is provided but it does not
            exist, or if the plan file does not exist.
        ValueError:
            Raised if all processing steps are turned off, if the
            selected MPL is undefined; if the fast node is selected with
            nodes > 1; if multiple run modes are selected; or if the
            plate list, ifudesign list, or mode list are not provided
            during a redo mode or if they are provided with any other
            mode.
    """
    def __init__(self, overwrite=None, console=None, quiet=False, print_version=False,
                 strictver=True, mplver=None, redux_path=None, dapver=None, analysis_path=None,
                 plan_file=None, platelist=None, ifudesignlist=None, combinatorics=False,
                 list_file=None, sres_ext=None, sres_fill=None, covar_ext=None,
                 use_platetargets=False, platetargets=None, on_disk=False, log=False,
                 dapproc=True, pltifu_plots=True, post_process=False, post_plots=False,
                 report_progress=False, verbose=0, label='mangadap', nodes=1, cpus=None, qos=None,
                 umask='0027', walltime='240:00:00', hard=True, create=False, submit=False,
                 queue=None):

        # Save run-mode options
        self.overwrite = overwrite
        self.quiet = quiet
        self.print_version = print_version

        # Override environment
        self.strictver = strictver
        self.mpl = mplver
        self.redux_path = None if redux_path is None else os.path.abspath(redux_path)
        self.dapver = dapver
        self.analysis_path = None if analysis_path is None else os.path.abspath(analysis_path)

        # List of files to analyze
        self.plan_file = None if plan_file is None else os.path.abspath(plan_file)
        if list_file is not None and not os.path.isfile(list_file):
            raise FileNotFoundError('No file: {0}'.format(list_file))
        if list_file is None:
            self.platelist = arginp_to_list(platelist, evaluate=True)
            self.ifudesignlist = arginp_to_list(ifudesignlist, evaluate=True)
        else:
            if platelist is not None or ifudesignlist is not None:
                warnings.warn('Provided file with list of files supercedes other input.')
            self.platelist, self.ifudesignlist = self._read_file_list(list_file)
        self.combinatorics = combinatorics
        self.sres_ext = sres_ext
        self.sres_fill = sres_fill
        self.covar_ext = covar_ext

        # Select if plateTargets files should be used to generate the
        # DRPComplete database, and possibly provide them directly 
        # TODO: This is likely now outdated
        self.use_platetargets = use_platetargets
        self.platetargets = arginp_to_list(platetargets)
        self.on_disk = on_disk

        # Set the options for output
        self.log = log
        self.dapproc = dapproc
        self.pltifu_plots = pltifu_plots
        self.post_process = post_process
        self.post_plots = post_plots

        # Report the progess of the cluster, otherwise just finish
        # script
        self.report_progress = report_progress
        self.verbose = verbose

        # Cluster queue keywords
        self.label = label
        self.nodes = nodes
        self.cpus = cpus
        self.qos = qos
        self.umask = umask
        self.walltime = walltime
        self.hard = hard
        self.submit = submit

        self.q = queue

        # Read and parse command-line arguments
        if console:
            self._read_arg()

        # Only print the version of the DAP
        if self.print_version:
            print('DAP Version: {0}'.format(__version__))
            return

        # If submitting to the cluster, make sure that scripts will be
        # created!
        if not self.create and self.submit:
            self.create = True

        # If setting qos, the number of nodes must be 1
        if self.qos is not None and self.nodes > 1:
            raise ValueError('When selecting the fast node, must have nodes=1!')

        # Check processing steps
        nproc = 0
        processes = []
        if self.dapproc:
            nproc += 1
            processes += [ 'Main DAP blocks' ]
        if self.pltifu_plots:
            nproc += 1
            processes += [ 'PLATEIFU-specific QA plots' ]
        if self.post_process:
            nproc += 1
            processes += [ 'DAPall' ]
        if self.post_plots:
            nproc += 1
            processes += [ 'DAPall and PLATE-specific QA plots' ]
        if nproc < 1:
            raise ValueError('No processing steps requested!')

        if (self.dapproc or self.pltifu_plots) and self.submit and self.post_process \
                and not self.report_progress:
            warnings.warn('If both processing and post processing, need to report progress.  '
                          'Setting to report progress.')
            self.report_progress = True

        # Make sure the selected MPL version is available
        try:
            self.mpl = MaNGAMPL(version=self.mpl, strictver=self.strictver)
        except Exception as e:
#            e = sys.exc_info()
#            print_frame(e[0])
#            raise ValueError('MPL is undefined: {0}'.format(e[1]))
            raise ValueError('MPL is undefined!') from e

        self.dapver = defaults.dap_version() if self.dapver is None else self.dapver

        # Set the output paths
        self.redux_path = defaults.drp_redux_path(self.mpl.drpver) if self.redux_path is None \
                                    else os.path.abspath(self.redux_path)
        self.analysis_path = defaults.dap_analysis_path(self.mpl.drpver, self.dapver) \
                                    if self.analysis_path is None \
                                    else os.path.abspath(self.analysis_path)

        # Set the subdirectory from which the scripts are sourced and
        # the various log files are written.  Currently based on start
        # UTC; e.g., analysis_path/log/19May2016T09.17.42UTC
        self.calling_path = os.path.join(self.analysis_path, 'log',
                                         time.strftime('%d%b%YT%H.%M.%SUTC',time.gmtime()))

        # Create the calling path. The existence test is executed, even
        # though time stamp should mean that calling_path never exists.
        # TODO: allow argument to set calling_path directly?
        if not os.path.isdir(self.calling_path):
            os.makedirs(self.calling_path)

        # Allow for the default plan.
        if self.plan_file is None:
            warnings.warn('No analysis plan was provided.  DAP executions will adopt the default '
                          'plan.')
        else:
            if not os.path.isfile(self.plan_file):
                raise FileNotFoundError('No file: {0}'.format(self.plan_file))
            _plan_file = self.plan_file
            self.plan_file = os.path.join(self.calling_path, os.path.split(self.plan_file)[1])
            if not os.path.isfile(self.plan_file):
                shutil.copy2(_plan_file, self.plan_file)

        # Construct the DAPTYPEs and check that the plans yeild a fully
        # unique list
        # TODO: This should probably be check done in AnalysisPlanSet
        self.plan = AnalysisPlanSet.default() if self.plan_file is None \
                        else AnalysisPlanSet.from_par_file(self.plan_file)
        self.daptypes = []
        for p in self.plan:
            bin_method = SpatiallyBinnedSpectra.define_method(p['bin_key'])
            sc_method = StellarContinuumModel.define_method(p['continuum_key'])
            el_method = None if p['elfit_key'] is None \
                                else EmissionLineModel.define_method(p['elfit_key'])
            self.daptypes += [defaults.dap_method(bin_method['key'],
                                                  sc_method['fitpar']['template_library_key'],
                                                  'None' if el_method is None else
                                                        el_method['continuum_tpl_key'])]
        if len(numpy.unique(self.daptypes)) != self.plan.nplans:
            raise ValueError('{0} do not yield unique DAPTYPEs.'.format(
                                'Default plans' if self.plan_file is None 
                                    else 'Plans in {0} '.format(self.plan_file)))

        # Alert the user of the versions to be used
        if self.q is not None:
            print('Attempting to submit to queue: {0}'.format(self.q))
        print('Versions: DAP:{0}, {1}'.format(self.dapver, self.mpl.mplver))
        self.mpl.show()
        self.mpl.verify_versions(quiet=self.quiet)
        print('Paths:')
        print('      REDUX: {0}'.format(self.redux_path))
        print('   ANALYSIS: {0}'.format(self.analysis_path))
        print('    CALLING: {0}'.format(self.calling_path))
        if self.plan_file is not None:
            print('  PLAN FILE: {0}'.format(self.plan_file))
        print('Processing steps added to scripts:')
        print('        {0}'.format(processes))

        # Create and update the DRPComplete file if necessary
        self.drpc = DRPComplete(platetargets=self.platetargets, drpver=self.mpl.drpver,
                                redux_path=self.redux_path, dapver=self.dapver,
                                analysis_path=self.analysis_path)
        
        # Update the DRPComplete list; force an update if platetarget
        # files are provided
        self.drpc.update(platelist=self.platelist, ifudesignlist=self.ifudesignlist,
                         combinatorics=self.combinatorics, force=(self.platetargets is not None),
                         use_platetargets=self.use_platetargets, on_disk=self.on_disk)

        # Hereafter, self.platelist and self.ifuplatelist are the INPUT
        # values, whereas self.drpc.platelist and self.drpc.ifuplatelist
        # are the actual DRP reductions to analyze (including
        # combinatorics and file existence tests).  See, e.g.,
        # selected_drpfile_list().

        # Find the rows in the DRPComplete database with the DRP file
        # meta data
        drpc_rows = self._selected_drpfile_rows()

        print('Number of DRP files to process: {0}'.format(len(drpc_rows)))
        if len(drpc_rows) == 0:
            # If coded correctly, this function should never return here.
            return

        # Construct the post-processing scripts before
        # constructing/submitting datacube processing scripts
        if self.post_process or self.post_plots:
            post_queue = self._build_post_queue(drpc_rows)

        # Process the relevant datacubes
        if self.dapproc or self.pltifu_plots:
            # Create the script files for the set of drpfiles, and
            # submit the scripts to the queue if self.submit is true
            proc_queue = self._build_proc_queue(drpc_rows)

            # Submit the queue to the cluster
            if self.create:
                proc_queue.commit(hard=self.hard, submit=self.submit)

            # If not reporting progress, done
            if not self.report_progress:
                return

            # Wait until the queue is finished
            running = True
            while running and self.submit:
                time.sleep(60) 
                percent_complete = proc_queue.get_percent_complete()
                print('Percent complete ... {0:.1f}%'.format(percent_complete), end='\r')
                if percent_complete == 100:
                    running = False
            print('Percent complete ... {0:.1f}%'.format(percent_complete))

        if not self.post_process and not self.post_plots:
            # Don't perform the post-processing
            return

        # Submit the post-processing queue to the cluster
        if self.create:
            post_queue.commit(hard=self.hard, submit=self.submit)

    # ******************************************************************
    #  UTILITY FUNCTIONS
    # ******************************************************************
    def _read_file_list(self, list_file):
        db = numpy.genfromtxt(list_file, dtype=object)
        if db.shape[1] != 2:
            raise ValueError('Input file should contain 2 columns, plate and ifu.')
        return db[:,0].astype(int).tolist(), db[:,1].astype(int).tolist()

    def _read_arg(self):
        """Read and interpret the terminal command-line arguments.

        Function uses the argparse module, which provide a --help option
        to list all available arguments.  The arguments mimic those
        required by the class constructor.
        """
        arg = parse_args()

        ################################################################
        # Assign properties based on arguments; the suitability of these
        # arguments is checked in __init__()

        # Run-mode options
        # Will OVERWRITE existing input from __init__()
        if arg.overwrite is not None:
            self.overwrite = arg.overwrite
        if arg.quiet is not None:
            self.quiet = arg.quiet
        if arg.print_version is not None:
            self.print_version = arg.print_version

        # Set the versions to use
        # Will OVERWRITE existing input from __init__()
        self.strictver = not arg.loose
        if arg.mplver is not None:
            self.mpl = arg.mplver
        if arg.redux_path is not None:
            self.redux_path = arg.redux_path
        if arg.dapver is not None:
            self.dapver = arg.dapver
        if arg.analysis_path is not None:
            self.analysis_path = arg.analysis_path

        if arg.plan_file is not None:
            self.plan_file = os.path.abspath(arg.plan_file)

        if arg.list_file is not None and not os.path.isfile(arg.list_file):
            raise FileNotFoundError('No file: {0}'.format(arg.list_file))

        if arg.list_file is None:
            if arg.platelist is not None:
                self.platelist = arginp_to_list(arg.platelist, evaluate=True)
            if arg.ifudesignlist is not None:
                self.ifudesignlist = arginp_to_list(arg.ifudesignlist, evaluate=True)
        else:
            if arg.platelist is not None or arg.ifudesignlist is not None:
                warnings.warn('Provided file with list of files supercedes other input.')
            self.platelist, self.ifudesignlist = self._read_file_list(arg.list_file)

        self.combinatorics = arg.combinatorics
        self.sres_ext = arg.sres_ext
        self.sres_fill = arg.sres_fill
        self.covar_ext = arg.covar_ext
   
        # Set the plateTargets and NSA catalog path
        if arg.plttargets is not None:
            self.platetargets = arginp_to_list(arg.plttargets)
        self.use_platetargets = arg.use_plttargets
        self.on_disk = arg.on_disk

        # Overwrite any existing output options
        self.log = arg.log
        self.dapproc = not arg.no_proc
        self.pltifu_plots = not arg.no_plots
        self.post_process = arg.post
        self.post_plots = arg.post_plots
        if self.verbose < arg.verbose:
            self.verbose = arg.verbose

        # Set to report the progress of the cluster.  This is needed if
        # waiting to submit the DAPall construction script
        self.report_progress = arg.progress

        # Set queue keywords
        if arg.label is not None:
            self.label = arg.label
        if arg.nodes is not None:
            self.nodes = arg.nodes
        if arg.cpus is not None:
            self.cpus = arg.cpus
        if arg.qos is not None:
            self.qos = arg.qos
            self.nodes = 1          # Force the number of nodes to be 1
        if arg.umask is not None:
            self.umask = arg.umask
        if arg.walltime is not None:
            self.walltime = arg.walltime
        if arg.hard is not None:
            self.hard = arg.hard
        if arg.create is not None:
            self.create = arg.create
        if arg.submit is not None:
            self.submit = arg.submit

        # Specify the destination queue
        if arg.queue is not None:
            self.q = arg.queue

    def _check_paths(self, plate, ifudesign):
        """
        Check if the output paths exists for a given plate and ifudesign.
        If not, create them.  The necessary output paths are:

            - The calling path that contains a copy of the plan, the
              script input and output files, and the touch files
            - The main common path defined by
              :func:`mangadap.config.defaults.dap_common_path`.
            - The plan-based subdirectories based on the plan file

        Args:
            plate (:obj:`int`):
                Plate number
            ifudesign (:obj:`int`):
                ifudesign number
        """
        # Generate the calling path for this plate/ifudesign
        path = os.path.join(self.calling_path, str(plate), str(ifudesign))
        if not os.path.isdir(path):
            os.makedirs(path)

        # Generate the main common path
        path = defaults.dap_common_path(plate=plate, ifudesign=ifudesign, drpver=self.mpl.drpver,
                                        dapver=self.dapver, analysis_path=self.analysis_path)
        if not os.path.isdir(path):
            os.makedirs(path)

        # Check the reference directories, which creates the full plan
        # path if necessary
        for daptype in self.daptypes:
            path = defaults.dap_method_path(daptype, plate=plate, ifudesign=ifudesign, ref=True,
                                            drpver=self.mpl.drpver, dapver=self.dapver,
                                            analysis_path=self.analysis_path)
            if not os.path.isdir(path):
                os.makedirs(path)


    def _create_queue(self):
        """
        Create a queue instance.
        
        This function is written to be functional for submission to the
        cluster queues both at Utah and using the sciama cluster at
        Portsmouth.  The available queues are:

        +-----------+--------+------+-----------+
        |     Queue |  Nodes |  PPN |   GB/core |
        +===========+========+======+===========+
        | sciama1.q |     80 |   12 |         2 |
        +-----------+--------+------+-----------+
        | sciama2.q |     96 |   16 |         4 |
        +-----------+--------+------+-----------+
        | sciama3.q |     48 |   20 |       3.2 |
        +-----------+--------+------+-----------+

        One should **not** submit to sciama3.q without permission from
        Will Percival.  The default queue is cluster.q (sciama2.q); so
        basically only use :attr:`q` to select sciama1.q.

        Submissions to Utah should have their queue destination set to
        None.  However, the fast node can be selected using :attr:`qos`.
        """
        # Instantiate
        queue = pbs.queue(verbose=not self.quiet)

        if self.q == 'ember':
            # Submitting to Utah ember cluster
            ppn = 12
            cpus = ppn if self.cpus is None else min(self.cpus, ppn)
            walltime = self.walltime if int(self.walltime.split(':')[0]) < 72 else '72:00:00'
            queue.create(label=self.label, nodes=self.nodes, qos=self.qos, umask=self.umask,
                         walltime=walltime, ppn=ppn, cpus=cpus, partition='ember', alloc='sdss')
        elif self.q is not None:
            # All other self.q values expected for Portsmouth cluster,
            # sciama.  In this case, the number of nodes is queue
            # dependent, and qos is not set
            if self.q == 'sciama1.q':
                ppn = 12
            elif self.q == 'sciama3.q':
                ppn = 20
            else:
                ppn = 16
            cpus = ppn if self.cpus is None else min(self.cpus, ppn)
            queue.create(label=self.label, nodes=self.nodes, umask=self.umask,
                         walltime=self.walltime, queue=self.q, ppn=ppn, cpus=cpus)
        else:
            # self.q can be None when submitting to both the Portsmouth
            # and Utah clusters.  In this case, the default queue
            # destination and ppn is correct.  qos is also set, but this
            # should only be used when submitting to Utah.
            ppn = 16
            cpus = ppn if self.cpus is None else min(self.cpus, ppn)
            queue.create(label=self.label, nodes=self.nodes, qos=self.qos, umask=self.umask,
                         walltime=self.walltime, ppn=ppn, cpus=cpus)

        return queue

    # Files:
    #       - mangadap-{plate}-{ifudesign}-LOG{mode} = script file = *
    #       - *.ready = script/directory is ready for queue submission
    #       - *.queued = script submitted to queue
    #       - *.started = script assigned a node/processor and has started running
    #       - *.done = completed execution of the full script
    # Other:
    #       - *.par = parameter file
    #       - *.out, *.err = stdout and stderr output from the script
    def _build_proc_queue(self, drpc_rows):
        """
        Create a queue instance, write the script files for all CPUs
        using :func:`prepare_for_analysis` and append the job to the
        queue, and then submit the jobs to the cluster queue.  If
        :attr:`submit` is False, this is essentially a wrapper for
        :func:`prepare_for_analysis`.

        This function is written to be functional for submission to the
        cluster queues both at Utah and using the sciama cluster at
        Portsmouth.  The available queues are:

        +-----------+--------+------+-----------+
        |     Queue |  Nodes |  PPN |   GB/core |
        +===========+========+======+===========+
        | sciama1.q |     80 |   12 |         2 |
        +-----------+--------+------+-----------+
        | sciama2.q |     96 |   16 |         4 |
        +-----------+--------+------+-----------+
        | sciama3.q |     48 |   20 |       3.2 |
        +-----------+--------+------+-----------+

        One should **not** submit to sciama3.q without permission from
        Will Percival.  The default queue is cluster.q (sciama2.q); so
        basically only use :attr:`q` to select sciama1.q.

        Submissions to Utah should have their queue destination set to
        None.  However, the fast node can be selected using :attr:`qos`.

        .. todo::
            - This algorithm is slow because it has to search through
              the drpcomplete fits file each time to find the
              plate/ifudesign combo.  Should make this more efficient.

        Args:
            drpc_rows (`numpy.ndarray`):
                A vector with the rows in the
                :class:`mangadap.survey.drpcomplete.DRPComplete`
                database with the DRP files to analyze.
        """
        # If submitting to the queue, create the queue object.  The
        # creation of the queue object is system dependent; however,
        # once created, the queue object is treated identically for both
        # the Utah and Portsmouth clusters.
        if self.create:
            queue = self._create_queue()
       
        # Create the script files, regardless of whether or not they are
        # submitted to the queue, appending the scripts to the queue if
        # they are to be submitted
        ndrp = len(drpc_rows)
        for i, index in enumerate(drpc_rows):
            # Write the compute script (write_compute_script also checks
            # the path exists!)
            scriptfile, stdoutfile, stderrfile \
                    = self.write_compute_script(index, dapproc=self.dapproc,
                                                plots=self.pltifu_plots, overwrite=self.overwrite)

            # Set the status to ready
            plate = self.drpc['PLATE'][index]
            ifudesign = self.drpc['IFUDESIGN'][index]
            self.set_pltifu_status(plate, ifudesign, status='ready')

            if self.create:
                queue.append('source {0}'.format(scriptfile), outfile=stdoutfile,
                                  errfile=stderrfile)
                self.set_pltifu_status(plate, ifudesign, status='queued')
            else:
                print('Preparing for analysis...{0:.1f}%'.format((i+1)*100/ndrp), end='\r')

        if not self.create:
            print('Preparing for analysis...{0:.1f}%'.format(100))

        return queue if self.create else None

    def _build_post_queue(self, drpc_rows):
        """
        Build the queue that performs the post-processing steps.
        """
        # Create the queue
        if self.create:
            queue = self._create_queue()

        if self.post_process:
            # Write the DAPall script and set its status
            dapall_scr, dapall_out, dapall_err \
                    = self.write_dapall_script(plots=self.post_plots, overwrite=self.overwrite)
            self.set_status(dapall_scr, status='ready')
            if self.create:
                # Add it to the queue
                queue.append('source {0}'.format(dapall_scr), outfile=dapall_out,
                             errfile=dapall_err)
                self.set_status(dapall_scr, status='queued')

        # Write the plate QA plot scripts
        if self.post_plots:
            # Find the unique plates
            plates = numpy.unique(self.drpc['PLATE'][drpc_rows])
            for plt in plates:
                # Write the plot script for this plate
                scriptfile, stdoutfile, stderrfile \
                        = self.write_plate_qa_script(plt, overwrite=self.overwrite)
                self.set_status(scriptfile, status='ready')
                
                if self.create:
                    queue.append('source {0}'.format(scriptfile), outfile=stdoutfile,
                                 errfile=stderrfile)
                    self.set_status(scriptfile, status='queued')

        return queue if self.create else None

    # ******************************************************************
    # Management
    # ******************************************************************
    def set_pltifu_status(self, plate, ifudesign, mode=None, status='queued'):
        """
        Generate a touch file that signifies the status for a given
        plate/ifudesign/mode.  The root of the touch file is set by
        :func:`mangadap.config.defaults.dap_file_root`.

        Args:
            plate (int): Plate number
            ifudesign (int): IFU design number
            mode (str): (**Optional**) DRP 3D mode, 'RSS' or 'CUBE'
            status (str): (**Optional**) That status signifier for the
                touch file.
        """
        # Get the name of the status file        
        root = defaults.dap_file_root(plate, ifudesign, mode=mode)
        self.set_status(os.path.join(self.calling_path, str(plate), str(ifudesign), root),
                        status)
        
    def set_status(self, root, status='queued'):
        """
        Generate a general touch file.

        Args:
            root (str): Full path and root name of the touch file
            status (str): (**Optional**) That status signifier for the
                touch file.
        """
        # Touch the file by opening and closing it
        file = open('{0}.{1}'.format(root, status), 'w')
        file.close()

    def _selected_drpfile_rows(self):
        """
        Find the rows in the
        :class:`mangadap.survey.drpcomplete.DRPComplete` database with
        the selected plates and IFUs to analyze.

        Returns:
            `numpy.ndarray`: A vector with the rows in the DRPComplete
            object with the relevant DRP file data.
        """
        # Find the rows in the DRPComplete database with the specified
        # plate and IFUS

        # All available PLATEIFUs
        drpc_pltifu = numpy.array(['{0}-{1}'.format(p,i) 
                                    for p,i in zip(self.drpc['PLATE'], self.drpc['IFUDESIGN'])])
        # The PLATEIFUs to analyze
        this_pltifu = numpy.array(['{0}-{1}'.format(p,i) 
                                    for p,i in zip(self.drpc.platelist, self.drpc.ifudesignlist)])

        # The rows with the PLATEIFUs to analyze
        rows = numpy.array([numpy.where(drpc_pltifu == pi)[0][0] if pi in drpc_pltifu else -1 \
                                for pi in this_pltifu])
        if numpy.any(rows < 0):
            raise ValueError('Should be able to find all plates and ifus.')
        return rows

    def write_compute_script(self, index, dapproc=True, plots=True, overwrite=False,
                              relative_symlink=True):
        """
        Write the MaNGA DAP script file that is sent to a single CPU to
        analyze a single DRP file with a given plate and ifudesign.

        Args:
            index (:obj:`int`, optional):
                Row in the
                :class:`mangadap.survey.drpcomplete.DRPComplete`
                database with the DRP file to analysis.
            dapproc (:obj:`bool`, optional):
                Flag to execute the main DAP processing.
            plots (:obj:`bool`, optional):
                Create the QA plots.
            overwrite (:obj:`bool`, optional):
                Flag to overwrite any existing files.
            relative_symlink (:obj:`bool`, optional):
                Set the symlink to the par file to be a relative path
                instead of the absolute path.

        Returns:
            :obj:`str`: Three strings with the name of the written
            script file, the file for the output sent to STDOUT, and the
            file for the output sent to STDERR.
        """
        # Get the plate and IFU from the DRPComplete database; alway use
        # the CUBE file
        plate = self.drpc['PLATE'][index]
        ifudesign = self.drpc['IFUDESIGN'][index]
        mode = 'CUBE'

        # Check that the path exists, creating it if not
        # TODO: More efficient way to do this?
        self._check_paths(plate, ifudesign)

        # Create the parameter file
#        parfile = defaults.dap_par_file(plate, ifudesign, mode, drpver=self.mpl.drpver,
#                                        dapver=self.dapver, analysis_path=self.analysis_path)
        cfgfile = defaults.dap_config(plate, ifudesign, drpver=self.mpl.drpver, dapver=self.dapver,
                                      analysis_path=self.analysis_path)

        # Write the par file if it doesn't exist
        if not os.path.isfile(cfgfile) or overwrite:
            # overwrite defaults to True
            self.drpc.write_config(cfgfile, index=index, sres_ext=self.sres_ext,
                                   sres_fill=self.sres_fill, covar_ext=self.covar_ext)
            # and create symlinks to it
            for daptype in self.daptypes:
                # Generate the ref subdirectory for this plan
                path = defaults.dap_method_path(daptype, plate=plate, ifudesign=ifudesign,
                                                ref=True, drpver=self.mpl.drpver,
                                                dapver=self.dapver,
                                                analysis_path=self.analysis_path)
                create_symlink(cfgfile, path, relative_symlink=relative_symlink,
                               overwrite=overwrite, quiet=True)

        # Set the root path for the scripts, inputs, outputs, and logs
        _calling_path = os.path.join(self.calling_path, str(plate), str(ifudesign))
        scr_file_root = defaults.dap_file_root(plate, ifudesign)

        # Set the names for the script, stdout, and stderr files
        scriptfile = os.path.join(_calling_path, scr_file_root)
        stdoutfile = '{0}.out'.format(scriptfile)
        stderrfile = '{0}.err'.format(scriptfile)

        ################################################################
        # Script file already exists, so just return
        if os.path.exists(scriptfile) and not overwrite:
            return scriptfile, stdoutfile, stderrfile

        # Main script components are:
        #   - (dapproc) Run the DAP processing code
        #   - (plots) Run the plotting scripts to plot the output
        # ONE OF THESE 2 MUST BE TRUE

        ################################################################
        # Open the script file and write the date as a commented header
        # line
        file = open(scriptfile, 'w')
        file.write('# Auto-generated batch file\n')
        file.write('# {0}\n'.format(time.strftime("%a %d %b %Y %H:%M:%S",time.localtime())))
        file.write('\n')

        # Create the started touch file
        startfile = '{0}.started'.format(scriptfile)
        file.write('touch {0}\n'.format(startfile))
        file.write('\n')

        # Command that runs the DAP
        # TODO: Define a "default" plan?
        if dapproc:
            command = 'OMP_NUM_THREADS=1 manga_dap -c {0} -r {1} -a {2}'.format(
                            cfgfile, self.redux_path, self.analysis_path)
            if self.plan_file is not None:
                command += ' -p {0}'.format(self.plan_file)
            if self.log:
                command += ' --log {0}.log'.format(scriptfile)
            if self.verbose > 0:
                command += (' -'+'v'*self.verbose)
            file.write('{0}\n'.format(command))
            file.write('\n')

        # Plotting scripts
        if plots:
            command = 'OMP_NUM_THREADS=1 dap_ppxffit_qa {0} {1} --analysis_path {2}'.format(
                            plate, ifudesign, self.analysis_path)
            if self.plan_file is not None:
                command += ' --plan_file {0}'.format(self.plan_file)
            file.write('{0}\n'.format(command))
            file.write('\n')

            command = 'OMP_NUM_THREADS=1 spotcheck_dap_maps {0} {1} --analysis_path {2}'.format(
                            plate, ifudesign, self.analysis_path)
            if self.plan_file is not None:
                command += ' --plan_file {0}'.format(self.plan_file)
            file.write('{0}\n'.format(command))
            file.write('\n')

            command = 'OMP_NUM_THREADS=1 dap_fit_residuals {0} {1} --analysis_path {2}'.format(
                            plate, ifudesign, self.analysis_path)
            if self.plan_file is not None:
                command += ' --plan_file {0}'.format(self.plan_file)
            file.write('{0}\n'.format(command))
            file.write('\n')

        # Touch the done file
        donefile = '{0}.done'.format(scriptfile)
        file.write('touch {0}\n'.format(donefile))
        file.write('\n')

        # Done
        file.close()

        # Return the script file, file for stdout, and file for stderr
        return scriptfile, stdoutfile, stderrfile
    
    def write_dapall_script(self, plots=True, overwrite=False):
        """
        Write the script used to construct the DAPall file and its
        associated quality assessment plots.

        Args:
            plots (:obj:`bool`, optional):
                Create the QA plots. Default is True.
            overwrite (:obj:`bool`, optional):
                Flag to overwrite any existing files.

        Returns:
            :obj:`str`: Three strings with the name of the written
            script file, the file for the output sent to STDOUT, and the
            file for the output sent to STDERR.
        """
        # Check that the path exists, creating it if not
        if not os.path.isdir(self.calling_path):
            os.makedirs(self.calling_path)
        
        # Set the names for the script, stdout, and stderr files
        scriptfile = os.path.join(self.calling_path, 'build_dapall')
        stdoutfile = '{0}.out'.format(scriptfile)
        stderrfile = '{0}.err'.format(scriptfile)

        # Script file already exists, so just return
        if os.path.exists(scriptfile) and not overwrite:
            return scriptfile, stdoutfile, stderrfile

        # Open the script file and write the date as a commented header
        # line
        file = open(scriptfile, 'w')
        file.write('# Auto-generated batch file\n')
        file.write('# {0}\n'.format(time.strftime("%a %d %b %Y %H:%M:%S",time.localtime())))
        file.write('\n')

        # Create the started touch file
        startfile = '{0}.started'.format(scriptfile)
        file.write('touch {0}\n'.format(startfile))
        file.write('\n')

        # Command that constructs the DAPall file
        command = 'OMP_NUM_THREADS=1 '
        command += 'construct_dapall --drpver {0} -r {1} --dapver {2} -a {3}'.format(
                        self.mpl.drpver, self.redux_path, self.dapver, self.analysis_path)
        if self.plan_file is not None:
            command += ' --plan_file {0}'.format(self.plan_file)
        if self.verbose > 0:
            command += (' -'+'v'*self.verbose )
        file.write('{0}\n'.format(command))
        file.write('\n')

        # Add the plotting commands
        if plots:
            command = 'OMP_NUM_THREADS=1 dapall_qa --drpver {0} --redux_path {1}'.format(
                            self.mpl.drpver, self.redux_path) \
                        + ' --dapver {0} --analysis_path {1}'.format(self.dapver,
                                                                     self.analysis_path)
            if self.plan_file is not None:
                command += ' --plan_file {0}'.format(self.plan_file)
            file.write('{0}\n'.format(command))
            file.write('\n')

        # Touch the done file
        donefile = '{0}.done'.format(scriptfile)
        file.write('touch {0}\n'.format(donefile))
        file.write('\n')

        file.close()
        ################################################################

        # Return the script file, file for stdout, and file for stderr
        return scriptfile, stdoutfile, stderrfile
    
    def write_plate_qa_script(self, plate, overwrite=False):
        """
        Write the script used to create the plate fit QA plot.

        Args:
            plate (:obj:`int`):
                The plate number for the plot.
            overwrite (:obj:`bool`, optional):
                Flag to overwrite any existing files.

        Returns:
            :obj:`str`: Three strings with the name of the written
            script file, the file for the output sent to STDOUT, and the
            file for the output sent to STDERR.
        """
        # Check that the path exists, creating it if not
        path = os.path.join(self.calling_path, str(plate))
        if not os.path.isdir(path):
            os.makedirs(path)
        
        # Set the names for the script, stdout, and stderr files
        scriptfile = os.path.join(self.calling_path, str(plate), '{0}_fitqa'.format(plate))
        stdoutfile = '{0}.out'.format(scriptfile)
        stderrfile = '{0}.err'.format(scriptfile)

        # Script file already exists, so just return
        if os.path.exists(scriptfile) and not overwrite:
            return scriptfile, stdoutfile, stderrfile

        # Open the script file and write the date as a commented header
        # line
        file = open(scriptfile, 'w')
        file.write('# Auto-generated batch file\n')
        file.write('# {0}\n'.format(time.strftime("%a %d %b %Y %H:%M:%S",time.localtime())))
        file.write('\n')

        # Create the started touch file
        startfile = '{0}.started'.format(scriptfile)
        file.write('touch {0}\n'.format(startfile))
        file.write('\n')

        # Command that creates the plots
        command = 'OMP_NUM_THREADS=1 dap_plate_fit_qa {0} --analysis_path {1}'.format(
                        plate, self.analysis_path)
        if self.plan_file is not None:
            command += ' --plan_file {0}'.format(self.plan_file)
        file.write('{0}\n'.format(command))
        file.write('\n')

        # Touch the done file
        donefile = '{0}.done'.format(scriptfile)
        file.write('touch {0}\n'.format(donefile))
        file.write('\n')

        file.close()
        ################################################################

        # Return the script file, file for stdout, and file for stderr
        return scriptfile, stdoutfile, stderrfile
    






import numpy
from astropy.io import fits
from matplotlib import pyplot

from mangadap.util.bitmask import BitMask
from mangadap.config import defaults

sdssbits_file = defaults.sdss_maskbits_file()

targ1bm = BitMask.from_par_file(sdssbits_file, 'MANGA_TARGET1')
targ2bm = BitMask.from_par_file(sdssbits_file, 'MANGA_TARGET2')
targ3bm = BitMask.from_par_file(sdssbits_file, 'MANGA_TARGET3')

hdu = fits.open('drpall-v3_1_1.fits')

#print(hdu['MANGA'].columns.names)

indx = hdu['MANGA'].data['z'] < 0

mngtarg1 = hdu['MANGA'].data['mngtarg1'][indx]
mngtarg2 = hdu['MANGA'].data['mngtarg2'][indx]
mngtarg3 = hdu['MANGA'].data['mngtarg3'][indx]

print('MANGA_TARGET1')
for b in numpy.unique(mngtarg1):
    print(b, targ1bm.flagged_bits(b))

print('MANGA_TARGET2')
for b in numpy.unique(mngtarg2):
    print(b, targ2bm.flagged_bits(b))

print('MANGA_TARGET3')
for b in numpy.unique(mngtarg3):
    print(b, targ3bm.flagged_bits(b))

indx = numpy.where(indx)[0]
for i in indx:
    print('{0:>5} {1:>5} {2:12.5f} {3:8.2f} {4:7.1f} {5:7.1f} {6:>15} {7:>15} {8:>15}'.format(
          hdu['MANGA'].data['plate'][i], hdu['MANGA'].data['ifudsgn'][i], 
          hdu['MANGA'].data['z'][i], hdu['MANGA'].data['nsa_elpetro_ba'][i], 
          hdu['MANGA'].data['nsa_elpetro_phi'][i], hdu['MANGA'].data['nsa_elpetro_th50_r'][i], 
          ','.join(targ1bm.flagged_bits(hdu['MANGA'].data['mngtarg1'][i])),
          ','.join(targ2bm.flagged_bits(hdu['MANGA'].data['mngtarg2'][i])),
          ','.join(targ3bm.flagged_bits(hdu['MANGA'].data['mngtarg3'][i]))))

print(numpy.sum(indx))
exit()


indx = numpy.where(hdu['MANGA'].data['PLATEIFU'] == '8261-12705')[0][0]

print(hdu['MANGA'].data['z'][indx])
print(hdu['MANGA'].data['nsa_sersic_ba'][indx], hdu['MANGA'].data['nsa_elpetro_ba'][indx])
print(hdu['MANGA'].data['nsa_sersic_phi'][indx], hdu['MANGA'].data['nsa_elpetro_phi'][indx])
print(hdu['MANGA'].data['nsa_sersic_th50'][indx], hdu['MANGA'].data['nsa_elpetro_th50_r'][indx])
exit()

pyplot.scatter(hdu['MANGA'].data['nsa_sersic_ba'], hdu['MANGA'].data['nsa_elpetro_ba'],
               marker='.', s=30, lw=0, color='k')
pyplot.show()






from .version import version

__version__ = version
__license__ = 'BSD3'
__author__ = 'Kyle B. Westfall'
__maintainer__ = 'Kyle B. Westfall'
__email__ = 'westfall@ucolick.org'
__copyright__ = '(c) 2014-2022, SDSS-IV/MaNGA Pipeline Group'
__credits__ = ['Kyle B. Westfall', 'Brett H. Andrews', 'Jorge Barrera-Ballesteros',
               'Francesco Belfiore', 'Matthew A. Bershady', 'Kevin Bundy', 'Joel R. Brownstein',
               'Michele Cappellari', 'Brian Cherinka', 'Lodovico Coccato', 'Niv Drory',
               'Cheng Du', 'Eric Emsellem', 'Daniel Goddard', 'Xihan Ji', 'David R. Law',
               'Niu Li', 'Claudia Maraston', 'Karen Masters', 'Héctor Javier Ibarra Medel',
               'Taniya Parikh', 'Sebastián F. Sánchez', 'José R. Sánchez-Gallego',
               'Adam Schaefer', 'Shravan Shetty', 'Daniel Thomas', 'Christy A. Tremonti',
               'Anne-Marie Weijmans', 'Renbin Yan', 'Meng Yang', 'Zheng Zheng', 'Shuang Zhou']

#def short_warning(message, category, filename, lineno, file=None, line=None):
#    """
#    Return the format for a short warning message.
#    """
#    return ' %s: %s\n' % (category.__name__, message)
#
#import warnings
#warnings.formatwarning = short_warning




from IPython import embed

import numpy
from astropy.io import fits

from mangadap.datacube import MaNGADataCube

from mangadap.par.artifactdb import ArtifactDB
from mangadap.par.emissionlinedb import EmissionLineDB

from mangadap.util.drpbitmask import DRPFitsBitMask
from mangadap.util.pixelmask import SpectralPixelMask

from mangadap.proc.templatelibrary import TemplateLibrary
from mangadap.proc.ppxffit import PPXFFit
from mangadap.proc.stellarcontinuummodel import StellarContinuumModelBitMask

from mangadap.tests.util import data_test_file


def test_ppxffit():
    # Read the data
    specfile = data_test_file('MaNGA_test_spectra.fits.gz')
    hdu = fits.open(specfile)
    drpbm = DRPFitsBitMask()
    flux = numpy.ma.MaskedArray(hdu['FLUX'].data, mask=drpbm.flagged(hdu['MASK'].data,
                                                                MaNGADataCube.do_not_fit_flags()))
    ferr = numpy.ma.power(hdu['IVAR'].data, -0.5)
    flux[ferr.mask] = numpy.ma.masked
    ferr[flux.mask] = numpy.ma.masked
    nspec = flux.shape[0]

    # Instantiate the template libary
    velscale_ratio = 4
    tpl = TemplateLibrary('MILESHC', match_resolution=False, velscale_ratio=velscale_ratio,
                          spectral_step=1e-4, log=True, hardcopy=False)
    tpl_sres = numpy.mean(tpl['SPECRES'].data, axis=0)

    # Get the pixel mask
    pixelmask = SpectralPixelMask(artdb=ArtifactDB.from_key('BADSKY'),
                                  emldb=EmissionLineDB.from_key('ELPSCMSK'))

    # Instantiate the fitting class
    ppxf = PPXFFit(StellarContinuumModelBitMask())

    # Perform the fit
    fit_wave, fit_flux, fit_mask, fit_par \
        = ppxf.fit(tpl['WAVE'].data.copy(), tpl['FLUX'].data.copy(), hdu['WAVE'].data, flux, ferr,
                   hdu['Z'].data, numpy.full(nspec, 100.), iteration_mode='no_global_wrej',
                   reject_boxcar=100, ensemble=False, velscale_ratio=velscale_ratio,
                   mask=pixelmask, matched_resolution=False, tpl_sres=tpl_sres,
                   obj_sres=hdu['SRES'].data, degree=8, moments=2)

    # Test the results

    # Rejected pixels
    assert numpy.sum(ppxf.bitmask.flagged(fit_mask, flag='PPXF_REJECT')) == 119, \
                'Different number of rejected pixels'

    # Unable to fit
    assert numpy.array_equal(ppxf.bitmask.flagged_bits(fit_par['MASK'][5]), ['NO_FIT']), \
                'Expected NO_FIT in 6th spectrum'

    # Number of used templates
    assert numpy.array_equal(numpy.sum(numpy.absolute(fit_par['TPLWGT']) > 1e-10, axis=1),
                             [12, 13, 17, 15, 15,  0,  8, 12]), \
                'Different number of templates with non-zero weights'

    # Number of additive coefficients
    assert fit_par['ADDCOEF'].shape[1] == 9, 'Incorrect number of additive coefficients'

    # No multiplicative coefficients
    assert numpy.all(fit_par['MULTCOEF'] == 0), \
                'No multiplicative coefficients should exist'

    # Kinematics and errors
    assert numpy.all(numpy.absolute(fit_par['KIN'] -
                        numpy.array([[ 14880.7, 292.9], [ 15053.4, 123.2],
                                     [ 14787.5, 236.4], [  8291.8, 169.7],
                                     [  9261.4, 202.7], [     0.0,   0.0],
                                     [  5123.5,  63.8], [  5455.6,  51.8]])) < 0.1), \
                'Kinematics are too different'

    assert numpy.all(numpy.absolute(fit_par['KINERR'] -
                        numpy.array([[2.0,1.9], [1.5,1.7], [ 2.4, 2.4], [2.2,2.3],
                                     [1.1,1.1], [0.0,0.0], [26.1,30.8], [4.7,7.5]])) < 0.1), \
                'Kinematic errors are too different'

    # Velocity dispersion corrections
    assert numpy.all(numpy.absolute(fit_par['SIGMACORR_SRES'] -
                        numpy.array([23.5, 10.1, 27.3, 38.7, 22.3,  0.0, 63.8, 23.8])) < 0.1), \
                'SRES corrections are too different'

    assert numpy.all(numpy.absolute(fit_par['SIGMACORR_EMP'] -
                        numpy.array([22.6,  0.0, 26.0, 38.2, 18.0,  0.0, 70.1,  0.0])) < 0.1), \
                'EMP corrections are too different'

    # Figures of merit
    assert numpy.all(numpy.absolute(fit_par['RCHI2'] -
                        numpy.array([ 1.94, 1.18, 1.40, 1.53, 2.50, 0.00, 1.06, 0.86])) < 0.01), \
                'Reduced chi-square too different'

    assert numpy.all(numpy.absolute(fit_par['RMS'] -
                        numpy.array([0.033, 0.019, 0.034, 0.023, 0.046, 0.000, 0.015, 0.015]))
                     < 0.001), 'RMS too different'

    assert numpy.all(numpy.absolute(fit_par['FRMS'] -
                        numpy.array([0.018, 0.023, 0.023, 0.032, 0.018, 0.000, 33.577, 0.148]))
                     < 0.001), 'Fractional RMS too different'

    assert numpy.all(numpy.absolute(fit_par['RMSGRW'][:,2] -
                        numpy.array([0.067, 0.037, 0.068, 0.046, 0.093, 0.000, 0.029, 0.027]))
                     < 0.001), 'Median absolute residual too different'




# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""
Logging routines.

----

.. include license and copyright
.. include:: ../include/copy.rst

----

.. include common links, assuming primary doc root is up one directory
.. include:: ../include/links.rst
"""
import logging
import warnings

from astropy.wcs import FITSFixedWarning

def init_DAP_logging(log, simple_warnings=True, append=False, keep_fits_warnings=False):
    """
    Initialize the logging preferences for the DAP.

    Args:
        log (:obj:`str`):
            File with log output.
        simple_warnings (:obj:`bool`, optional):
            Shorten warning messages.
        append (:obj:`bool`, optional):
            Append new log messages to existing log file; if False, file
            is overwritten.
        keep_fits_warning (:obj:`bool`, optional):
            Flag to not ignore FITSFixedWarning messages.

    .. todo::
        Use a file with the logging configuration.  See:
        https://docs.python.org/3.5/howto/logging.html#handler-basic

    """

    # Remove existing root logging
    if len(logging.root.handlers) > 0:
        for handler in logging.root.handlers:
            logging.root.removeHandler(handler)

    # Init global logging
    logging.basicConfig(level=logging.INFO,
                        format='%(levelname)-8s ::  %(message)s',
                        datefmt='%Y-%m-%d %H:%M:%S')

    # Set warnings format
    warnings.simplefilter('always')
    if not keep_fits_warnings:
        warnings.simplefilter('ignore', FITSFixedWarning)
    warnings.filterwarnings('ignore', 'unclosed file')
    logging.captureWarnings(True)
    if simple_warnings:
        warnings.formatwarning = short_warning
#    warnings.simplefilter('error', FutureWarning)
#    import numpy
#    warnings.simplefilter('error', numpy.VisibleDeprecationWarning)
#    warnings.simplefilter('error', SyntaxWarning)
#    warnings.simplefilter('error', DeprecationWarning)

    # Add file handler if wanted
    if log is not None:
        logfile = logging.FileHandler(log, mode=('a' if append else 'w'))
        logfile.setFormatter(
                logging.Formatter('%(asctime)s %(name)-10s %(levelname)-8s ::  %(message)s'))
        logfile.setLevel(logging.DEBUG)
        logging.getLogger('').addHandler(logfile)


def module_logging(name, verbose):
    """
    Return a number of `logging.Logger`_ objects, one for each verbosity
    level.
    """
    if verbose == 0:
        return None
    loggers = []
    for i in range(verbose):
        loggers += [ logging.getLogger('{0} {1}'.format(name,i+1)) ]
    return loggers
    
    
def log_output(loggers, v, lvl, message):
    """
    Write message to the log.

    Args:
        loggers (`logging.Logger`_):  Objects collecting the log for a
            given run of the DAP.
        v (int): Verbosity level of message.
        lvl (`logging.level`_):  Logging level
        message (str): Message to log.
    """
    if loggers is None:
        print(message)
        return

    if len(loggers) < v:
        return

    loggers[v-1].log(lvl,message)


def short_warning(message, category, filename, lineno, file=None, line=None):
    """
    Return the format for a short warning message.
    """
    return ' %s: %s' % (category.__name__, message)
    





# Imports
import time

from IPython import embed

import numpy

from astropy.io import fits
import astropy.constants
from matplotlib import pyplot

from mangadap.tests.util import drp_test_version

from mangadap.datacube import MaNGADataCube
from mangadap.config import manga
from mangadap.config import defaults
from mangadap.util.sampling import spectral_coordinate_step

from mangadap.util.resolution import SpectralResolution
from mangadap.util.pixelmask import SpectralPixelMask

from mangadap.par.artifactdb import ArtifactDB
from mangadap.par.emissionmomentsdb import EmissionMomentsDB
from mangadap.par.emissionlinedb import EmissionLineDB
from mangadap.par.absorptionindexdb import AbsorptionIndexDB
from mangadap.par.bandheadindexdb import BandheadIndexDB

from mangadap.proc.templatelibrary import TemplateLibrary
from mangadap.proc.emissionlinemoments import EmissionLineMoments
from mangadap.proc.sasuke import Sasuke
from mangadap.proc.ppxffit import PPXFFit
from mangadap.proc.stellarcontinuummodel import StellarContinuumModel, StellarContinuumModelBitMask
from mangadap.proc.emissionlinemodel import EmissionLineModelBitMask
from mangadap.proc.spectralfitting import EmissionLineFit
from mangadap.proc.spectralindices import SpectralIndices

#-----------------------------------------------------------------------------

def get_redshift(plt, ifu, drpall_file=None):
    """
    Get the redshift of a galaxy from the DRPall file.

    Args:
        plt (:obj:`int`):
            Plate number
        ifu (:obj:`int`):
            IFU identifier
        drapall_file (:obj:`str`, optional):
            DRPall file. If None, attempts to use the default path to
            the file using environmental variables.
    
    Returns:
        :obj:`float`: The redshift to the galaxy observed by the
        provided PLATEIFU.
    """
    if drpall_file is None:
        drpall_file = manga.drpall_file()
    if not drpall_file.exists():
        raise FileNotFoundError(f'Could not find DRPall file: {drpall_file}')
    hdu = fits.open(str(drpall_file))
    indx = hdu[1].data['PLATEIFU'] == '{0}-{1}'.format(plt, ifu)
    return hdu[1].data['NSA_Z'][indx][0]


def get_spectra(plt, ifu, x, y, directory_path=None):
    """
    Extract spectra from a MaNGA observation.

    Args:
        plt (:obj:`int`):
            Plate number
        ifu (:obj:`int`):
            IFU identifier
        x (:obj:`int`, `numpy.ndarray`_):
            The spaxel coordinate along the RA axis.
        y (:obj:`int`, `numpy.ndarray`_):
            The spaxel coordinate along the DEC axis.
        directory_path (:obj:`str`, optional):
            Directory with the DRP LOGCUBE file. If None, uses the
            default directory path based on the environmental
            variables.

    Returns:
        :obj:`tuple`: Returns 4 numpy vectors: The wavelength, flux,
        flux inverse variance, and spectral resolution extracted from
        the datacube.
    """
    cube = MaNGADataCube.from_plateifu(plt, ifu, directory_path=directory_path)
    flat_indx = cube.spatial_shape[1]*x+y
    # This function always returns as masked array
    flux = cube.copy_to_masked_array(attr='flux', flag=cube.do_not_fit_flags())
    ivar = cube.copy_to_masked_array(attr='ivar', flag=cube.do_not_fit_flags())
    sres = cube.copy_to_array(attr='sres')
    return cube.wave, flux[flat_indx,:], ivar[flat_indx,:], sres[flat_indx,:]


#-----------------------------------------------------------------------------
def main():

    t = time.perf_counter()

    #-------------------------------------------------------------------
    # Read spectra to fit. The following reads a single MaNGA spectrum.
    # This is where you should read in your own spectrum to fit.
    # Plate-IFU to use
    plt = 7815
    ifu = 3702
    # Spaxel coordinates
    x = 25 #30
    y = 25 #37
    # Where to find the relevant datacube.  This example accesses the test data
    # that can be downloaded by executing the script here:
    # https://github.com/sdss/mangadap/blob/master/download_test_data.py
    directory_path = defaults.dap_source_dir() / 'data' / 'remote'
    # Read a spectrum
    wave, flux, ivar, sres = get_spectra(plt, ifu, x, y, directory_path=directory_path)
    # In general, the DAP fitting functions expect data to be in 2D
    # arrays with shape (N-spectra,N-wave). So if you only have one
    # spectrum, you need to expand the dimensions:
    flux = flux.reshape(1,-1)
    ivar = ivar.reshape(1,-1)
    ferr = numpy.ma.power(ivar, -0.5)
    sres = sres.reshape(1,-1)

    # The majority (if not all) of the DAP methods expect that your
    # spectra are binned logarithmically in wavelength (primarily
    # because this is what pPXF expects). You can either have the DAP
    # function determine this value (commented line below) or set it
    # directly. The value is used to resample the template spectra to
    # match the sampling of the spectra to fit (up to some integer; see
    # velscale_ratio).
#    spectral_step = spectral_coordinate_step(wave, log=True)
    spectral_step = 1e-4

    # Hereafter, the methods expect a wavelength vector, a flux array
    # with the spectra to fit, an ferr array with the 1-sigma errors in
    # the flux, and sres with the wavelength-dependent spectral
    # resolution, R = lambda / Dlambda
    #-------------------------------------------------------------------

    #-------------------------------------------------------------------
    # The DAP needs a reasonable guess of the redshift of the spectrum
    # (within +/- 2000 km/s). In this example, I'm pulling the redshift
    # from the DRPall file. There must be one redshift estimate per
    # spectrum to fit.  Here that means it's a single element array
    # This example accesses the test data
    # that can be downloaded by executing the script here:
    # https://github.com/sdss/mangadap/blob/master/download_test_data.py
    drpall_file = directory_path / f'drpall-{drp_test_version}.fits'
    z = numpy.array([get_redshift(plt, ifu, drpall_file)])
    print('Redshift: {0}'.format(z[0]))
    # The DAP also requires an initial guess for the velocity
    # dispersion. A guess of 100 km/s is usually robust, but this may
    # depend on your spectral resolution.
    dispersion = numpy.array([100.])
    #-------------------------------------------------------------------

    #-------------------------------------------------------------------
    # The following sets the keyword for the template spectra to use
    # during the fit. You can specify different template sets to use
    # during the stellar-continuum (stellar kinematics) fit and the
    # emission-line modeling.

    # Templates used in the stellar continuum fits
    sc_tpl_key = 'MILESHC'
    # Templates used in the emission-line modeling
    el_tpl_key = 'MASTARSSP'

    # You also need to specify the sampling for the template spectra.
    # The templates must be sampled with the same pixel step as the
    # spectra to be fit, up to an integer factor. The critical thing
    # for the sampling is that you do not want to undersample the
    # spectral resolution element of the template spectra. Here, I set
    # the sampling for the MILES templates to be a factor of 4 smaller
    # than the MaNGA spectrum to be fit (which is a bit of overkill
    # given the resolution difference). I set the sampling of the
    # MaStar templates to be the same as the galaxy data.

    # Template pixel scale a factor of 4 smaller than galaxy data
    sc_velscale_ratio = 4
    # Template sampling is the same as the galaxy data
    el_velscale_ratio = 1

    # You then need to identify the database that defines the
    # emission-line passbands (elmom_key) for the non-parametric
    # emission-line moment calculations, and the emission-line
    # parameters (elfit_key) for the Gaussian emission-line modeling.
    # See
    # https://sdss-mangadap.readthedocs.io/en/latest/emissionlines.html.
    elmom_key = 'ELBMPL9'
    elfit_key = 'ELPMPL11'

    # If you want to also calculate the spectral indices, you can
    # provide a keyword that indicates the database with the passband
    # definitions for both the absorption-line and bandhead/color
    # indices to measure. The script allows these to be None, if you
    # don't want to calculate the spectral indices. See
    # https://sdss-mangadap.readthedocs.io/en/latest/spectralindices.html
    absindx_key = 'EXTINDX'
    bhdindx_key = 'BHBASIC'

    # Now we want to construct a pixel mask that excludes regions with
    # known artifacts and emission lines. The 'BADSKY' artifact
    # database only masks the 5577, which can have strong left-over
    # residuals after sky-subtraction. The list of emission lines (set
    # by the ELPMPL8 keyword) can be different from the list of
    # emission lines fit below.
    sc_pixel_mask = SpectralPixelMask(artdb=ArtifactDB.from_key('BADSKY'),
                                      emldb=EmissionLineDB.from_key('ELPMPL11'))
    # Mask the 5577 sky line
    el_pixel_mask = SpectralPixelMask(artdb=ArtifactDB.from_key('BADSKY'))

    # Finally, you can set whether or not to show a set of plots.
    #
    # Show the ppxf-generated plots for each fit stage.
    fit_plots = False
    # Show summary plots
    usr_plots = True
    #-------------------------------------------------------------------


    #-------------------------------------------------------------------
    # Fit the stellar continuum

    # First, we construct the template library. The keyword that
    # selects the template library (sc_tpl_key) is defined above. The
    # following call reads in the template library and processes the
    # data to have the appropriate pixel sampling. Note that *no*
    # matching of the spectral resolution to the galaxy spectrum is
    # performed.
    sc_tpl = TemplateLibrary(sc_tpl_key, match_resolution=False, velscale_ratio=sc_velscale_ratio,
                             spectral_step=spectral_step, log=True, hardcopy=False)

    # This calculation of the mean spectral resolution is a kludge. The
    # template library should provide spectra that are *all* at the
    # same spectral resolution. Otherwise, one cannot freely combine
    # the spectra to fit the Doppler broadening of the galaxy spectrum
    # in a robust (constrained) way (without substantially more
    # effort). There should be no difference between what's done below
    # and simply taking the spectral resolution to be that of the first
    # template spectrum (i.e., sc_tpl['SPECRES'].data[0])
    sc_tpl_sres = numpy.mean(sc_tpl['SPECRES'].data, axis=0).ravel()

    # Instantiate the fitting class, including the mask that it should
    # use to flag the data. [[This mask should just be default...]]
    ppxf = PPXFFit(StellarContinuumModelBitMask())

    # The following call performs the fit to the spectrum. Specifically
    # note that the code only fits the first two moments, uses an
    # 8th-order additive polynomial, and uses the 'no_global_wrej'
    # iteration mode. See
    # https://sdss-mangadap.readthedocs.io/en/latest/api/mangadap.proc.ppxffit.html#mangadap.proc.ppxffit.PPXFFit.fit
    cont_wave, cont_flux, cont_mask, cont_par \
        = ppxf.fit(sc_tpl['WAVE'].data.copy(), sc_tpl['FLUX'].data.copy(), wave, flux, ferr,
                   z, dispersion, iteration_mode='no_global_wrej', reject_boxcar=100,
                   ensemble=False, velscale_ratio=sc_velscale_ratio, mask=sc_pixel_mask,
                   matched_resolution=False, tpl_sres=sc_tpl_sres, obj_sres=sres, degree=8,
                   moments=2, plot=fit_plots)

    # The returned objects from the fit are the wavelength, model, and
    # mask vectors and the record array with the best-fitting model
    # parameters. The datamodel of the best-fitting model parameters is
    # set by:
    # https://sdss-mangadap.readthedocs.io/en/latest/api/mangadap.proc.spectralfitting.html#mangadap.proc.spectralfitting.StellarKinematicsFit._per_stellar_kinematics_dtype

    # Remask the continuum fit
    sc_continuum = StellarContinuumModel.reset_continuum_mask_window(
                        numpy.ma.MaskedArray(cont_flux, mask=cont_mask>0))

    # Show the fit and residual
    if usr_plots:
        pyplot.plot(wave, flux[0,:], label='Data')
        pyplot.plot(wave, sc_continuum[0,:], label='Model')
        pyplot.plot(wave, flux[0,:] - sc_continuum[0,:], label='Resid')
        pyplot.legend()
        pyplot.xlabel('Wavelength')
        pyplot.ylabel('Flux')
        pyplot.show()
    #-------------------------------------------------------------------

    #-------------------------------------------------------------------
    # Get the emission-line moments using the fitted stellar continuum

    # Read the database that define the emission lines and passbands
    momdb = EmissionMomentsDB.from_key(elmom_key)

    # Measure the moments
    elmom = EmissionLineMoments.measure_moments(momdb, wave, flux, continuum=sc_continuum,
                                                redshift=z)
    #-------------------------------------------------------------------

    #-------------------------------------------------------------------
    # Fit the emission-line model

    # Set the emission-line continuum templates if different from those
    # used for the stellar continuum
    if sc_tpl_key == el_tpl_key:
        # If the keywords are the same, just copy over the previous
        # library ...
        el_tpl = sc_tpl
        el_tpl_sres = sc_tpl_sres
        # ... and the best fitting stellar kinematics
        stellar_kinematics = cont_par['KIN']
    else:
        # If the template sets are different, we need to match the
        # spectral resolution to the galaxy data ...
        _sres = SpectralResolution(wave, sres[0,:], log10=True)
        el_tpl = TemplateLibrary(el_tpl_key, sres=_sres, velscale_ratio=el_velscale_ratio,
                                 spectral_step=spectral_step, log=True, hardcopy=False)
        el_tpl_sres = numpy.mean(el_tpl['SPECRES'].data, axis=0).ravel()
        # ... and use the corrected velocity dispersions.
        stellar_kinematics = cont_par['KIN']
        stellar_kinematics[:,1] = numpy.ma.sqrt(numpy.square(cont_par['KIN'][:,1]) -
                                                    numpy.square(cont_par['SIGMACORR_EMP']))

    # Read the emission line fitting database
    emldb = EmissionLineDB.from_key(elfit_key)

    # Instantiate the fitting class
    emlfit = Sasuke(EmissionLineModelBitMask())

    # Perform the fit
    efit_t = time.perf_counter()
    eml_wave, model_flux, eml_flux, eml_mask, eml_fit_par, eml_eml_par \
            = emlfit.fit(emldb, wave, flux, obj_ferr=ferr, obj_mask=el_pixel_mask, obj_sres=sres,
                         guess_redshift=z, guess_dispersion=dispersion, reject_boxcar=101,
                         stpl_wave=el_tpl['WAVE'].data, stpl_flux=el_tpl['FLUX'].data,
                         stpl_sres=el_tpl_sres, stellar_kinematics=stellar_kinematics,
                         etpl_sinst_mode='offset', etpl_sinst_min=10.,
                         velscale_ratio=el_velscale_ratio, matched_resolution=False, mdegree=8,
                         plot=fit_plots)
    print('TIME: ', time.perf_counter() - efit_t)

    # Line-fit metrics
    eml_eml_par = EmissionLineFit.line_metrics(emldb, wave, flux, ferr, model_flux, eml_eml_par,
                                               model_mask=eml_mask, bitmask=emlfit.bitmask)

    # Get the stellar continuum that was fit for the emission lines
    elcmask = eml_mask.ravel() > 0
    goodpix = numpy.arange(elcmask.size)[numpy.invert(elcmask)]
    start, end = goodpix[0], goodpix[-1]+1
    elcmask[start:end] = False
    el_continuum = numpy.ma.MaskedArray(model_flux - eml_flux,
                                        mask=elcmask.reshape(model_flux.shape))

    # Plot the result
    if usr_plots:
        pyplot.plot(wave, flux[0,:], label='Data')
        pyplot.plot(wave, model_flux[0,:], label='Model')
        pyplot.plot(wave, el_continuum[0,:], label='EL Cont.')
        pyplot.plot(wave, sc_continuum[0,:], label='SC Cont.')
        pyplot.legend()
        pyplot.xlabel('Wavelength')
        pyplot.ylabel('Flux')
        pyplot.show()

    # Remeasure the emission-line moments with the new continuum
    new_elmom = EmissionLineMoments.measure_moments(momdb, wave, flux, continuum=el_continuum,
                                                    redshift=z)

    # Compare the summed flux and Gaussian-fitted flux for all the
    # fitted lines
    if usr_plots:
        pyplot.scatter(emldb['restwave'], (new_elmom['FLUX']-eml_eml_par['FLUX']).ravel(),
                       c=eml_eml_par['FLUX'].ravel(), cmap='viridis', marker='.', s=60, lw=0,
                       zorder=4)
        pyplot.grid()
        pyplot.xlabel('Wavelength')
        pyplot.ylabel('Summed-Gaussian Difference')
        pyplot.show()
    #-------------------------------------------------------------------

    #-------------------------------------------------------------------
    # Measure the spectral indices
    if absindx_key is None or bhdindx_key is None:
        # Neither are defined, so we're done
        print('Elapsed time: {0} seconds'.format(time.perf_counter() - t))
        return

    # Setup the databases that define the indices to measure
    absdb = None if absindx_key is None else AbsorptionIndexDB.from_key(absindx_key)
    bhddb = None if bhdindx_key is None else BandheadIndexDB.from_key(bhdindx_key)

    # Remove the modeled emission lines from the spectra
    flux_noeml = flux - eml_flux
    redshift = stellar_kinematics[:,0] / astropy.constants.c.to('km/s').value
    sp_indices = SpectralIndices.measure_indices(absdb, bhddb, wave, flux_noeml, ivar=ivar,
                                                 redshift=redshift)

    # Calculate the velocity dispersion corrections
    #   - Construct versions of the best-fitting model spectra with and without
    #     the included dispersion
    continuum = Sasuke.construct_continuum_models(emldb, el_tpl['WAVE'].data, el_tpl['FLUX'].data,
                                                  wave, flux.shape, eml_fit_par)
    continuum_dcnvlv = Sasuke.construct_continuum_models(emldb, el_tpl['WAVE'].data,
                                                         el_tpl['FLUX'].data, wave, flux.shape,
                                                         eml_fit_par, redshift_only=True)

    #   - Get the dispersion corrections and fill the relevant columns of the
    #     index table
    sp_indices['BCONT_MOD'], sp_indices['BCONT_CORR'], sp_indices['RCONT_MOD'], \
        sp_indices['RCONT_CORR'], sp_indices['MCONT_MOD'], sp_indices['MCONT_CORR'], \
        sp_indices['AWGT_MOD'], sp_indices['AWGT_CORR'], \
        sp_indices['INDX_MOD'], sp_indices['INDX_CORR'], \
        sp_indices['INDX_BF_MOD'], sp_indices['INDX_BF_CORR'], \
        good_les, good_ang, good_mag, is_abs \
                = SpectralIndices.calculate_dispersion_corrections(absdb, bhddb, wave, flux,
                                                                   continuum, continuum_dcnvlv,
                                                                   redshift=redshift,
                                                                   redshift_dcnvlv=redshift)

    # Apply the index corrections.  This is only done here for the
    # Worthey/Trager definition of the indices, as an example
    corrected_indices = numpy.zeros(sp_indices['INDX'].shape, dtype=float)
    corrected_indices_err = numpy.zeros(sp_indices['INDX'].shape, dtype=float)
    # Unitless indices
    corrected_indices[good_les], corrected_indices_err[good_les] \
            = SpectralIndices.apply_dispersion_corrections(sp_indices['INDX'][good_les],
                                                           sp_indices['INDX_CORR'][good_les],
                                                           err=sp_indices['INDX_ERR'][good_les])
    # Indices in angstroms
    corrected_indices[good_ang], corrected_indices_err[good_ang] \
            = SpectralIndices.apply_dispersion_corrections(sp_indices['INDX'][good_ang],
                                                           sp_indices['INDX_CORR'][good_ang],
                                                           err=sp_indices['INDX_ERR'][good_ang],
                                                           unit='ang')
    # Indices in magnitudes
    corrected_indices[good_mag], corrected_indices_err[good_mag] \
            = SpectralIndices.apply_dispersion_corrections(sp_indices['INDX'][good_mag],
                                                           sp_indices['INDX_CORR'][good_mag],
                                                           err=sp_indices['INDX_ERR'][good_mag],
                                                           unit='mag')

    # Print the results for a few indices
    index_names = numpy.append(absdb['name'], bhddb['name'])
    print('-'*73)
    print(f'{"NAME":<8} {"Raw Index":>12} {"err":>12} {"Index Corr":>12} {"Index":>12} {"err":>12}')
    print(f'{"-"*8:<8} {"-"*12:<12} {"-"*12:<12} {"-"*12:<12} {"-"*12:<12} {"-"*12:<12}')
    for name in ['Hb', 'HDeltaA', 'Mgb', 'Dn4000']:
        i = numpy.where(index_names == name)[0][0]
        print(f'{name:<8} {sp_indices["INDX"][0,i]:12.4f} {sp_indices["INDX_ERR"][0,i]:12.4f} '
              f'{sp_indices["INDX_CORR"][0,i]:12.4f} {corrected_indices[0,i]:12.4f} '
              f'{corrected_indices_err[0,i]:12.4f}')
    print('-'*73)

    embed()

    print('Elapsed time: {0} seconds'.format(time.perf_counter() - t))


if __name__ == '__main__':
    main()







from IPython import embed

from mangadap.scripts import scriptbase


class MangaDap(scriptbase.ScriptBase):

    @classmethod
    def name(cls):
        """
        Return the name of the executable.
        """
        return 'manga_dap'

    @classmethod
    def get_parser(cls, width=None):

        parser = super().get_parser(description='Perform analysis of integral-field data.',
                                    width=width)

        group = parser.add_mutually_exclusive_group(required=True)
        group.add_argument('-c', '--config', type=str, default=None,
                           help='Configuration file used to instantiate the relevant DataCube '
                                'derived class.')
        group.add_argument('-f', '--cubefile', type=str, default=None,
                           help='Name of the file with the datacube data.  Must be possible to '
                                'instantiate the relevant DataCube derived class directly from '
                                'the file only.')
        parser.add_argument('--cube_module', nargs='*',
                            default='mangadap.datacube.MaNGADataCube',
                            help='The name of the module that contains the DataCube derived '
                                 'class used to read the data.')

        parser.add_argument('-p', '--plan', type=str,
                            help='TOML file with analysis plan.  If not provided, a '
                                 'default plan is used.')
        parser.add_argument('--plan_module', nargs='*',
                            default='mangadap.config.manga.MaNGAAnalysisPlan',
                            help='The name of the module used to define the analysis plan and '
                                 'the output paths.')

        parser.add_argument('--dbg', help='Run manga_dap in debug mode', action='store_true',
                            default=False)
        parser.add_argument('--log', type=str, help='File name for runtime log', default=None)
        parser.add_argument('-v', '--verbose', action='count', default=0,
                            help='Set verbosity level; can be omitted and set up to -vv')

        parser.add_argument('-o', '--output_path', type=str, default=None,
                            help='Top-level directory for the DAP output files; default path is '
                                 'set by the provided analysis plan object (see plan_module).')
        return parser

    @staticmethod
    def main(args):

#        import warnings
#        warnings.simplefilter('error', DeprecationWarning)
        from mangadap.config.analysisplan import AnalysisPlan
        from mangadap.survey.manga_dap import manga_dap
        from mangadap.datacube import DataCube
        from mangadap.util.pkg import load_object

        # Execute the MaNGA DAP for a single datacube

        #   - Import the module used to read the datacube
        if isinstance(args.cube_module, list) and len(args.cube_module) > 2:
            raise ValueError('Provided cube module must be one or two strings.')
        if isinstance(args.cube_module, str) or len(args.cube_module) == 1:
            UserDataCube = load_object(args.cube_module if isinstance(args.cube_module, str) 
                                       else args.cube_module[0])
        else:
            UserDataCube = load_object(args.cube_module[0], obj=args.cube_module[1])
        #   - Check that the class is derived from DataCube
        if not issubclass(UserDataCube, DataCube):
            raise TypeError('Defined cube object must subclass from mangadap.datacube.DataCube.')

        #   - Import the module used to set the analysis plan
        if isinstance(args.plan_module, list) and len(args.plan_module) > 2:
            raise ValueError('Provided plan module must be one or two strings.')
        if isinstance(args.plan_module, str) or len(args.plan_module) == 1:
            UserPlan = load_object(args.plan_module if isinstance(args.plan_module, str) 
                                       else args.plan_module[0])
        else:
            UserPlan = load_object(args.plan_module[0], obj=args.plan_module[1])
        #   - Check that the class is derived from AnalysisPlan
        if not issubclass(UserPlan, AnalysisPlan):
            raise TypeError('Defined plan object must subclass from '
                            'mangadap.config.analysisplan.AnalysisPlan')

        #   - Instantiate the datacube object using either the datacube file
        #     directly or a configuration file
        cube = UserDataCube(args.cubefile) if args.config is None \
                    else UserDataCube.from_config(args.config)

        #   - Read the analysis plan
        plan = UserPlan.default(cube=cube, analysis_path=args.output_path) if args.plan is None \
                    else UserPlan.from_toml(args.plan, cube=cube, analysis_path=args.output_path)

        #   - Run the pipeline
        status = manga_dap(cube, plan, dbg=args.dbg, log=args.log, verbose=args.verbose)






from IPython import embed

from mangadap.scripts import scriptbase


# TODO: Move this somewhere that can be more generally accessed?
def calculate_covariance_cube(plate, ifudesign, ofile, nc=1, wave=None, directory_path=None):

    import numpy
    from mangadap.spectra import MaNGARSS

    print('     PLATE: {0}'.format(plate))
    print(' IFUDESIGN: {0}'.format(ifudesign))

    # Access the DRP RSS file
    print('Attempting to open RSS file:')
    rss = MaNGARSS.from_plateifu(plate, ifudesign, directory_path=directory_path)
    print(f'     FOUND: {rss.file_path}')

    if wave is not None:
        channel = numpy.argsort(numpy.absolute(rss.wave - wave))[0]
        print('Nearest wavelength channel has wavelength {0:.1f} ang.'.format(rss.wave[channel]))
        C = rss.covariance_matrix(channel)
    else:
        if nc >= rss.nwave or nc == 0:
            print('Calculating full covariance cube ...')
            C = rss.covariance_cube()
            print('... done.')
        elif nc == 1:
            channel = rss.nwave//2
            print('Calculating covariance matrix at central channel: '
                  '{0:.1f} ang.'.format(rss.wave[channel]))
            C = rss.covariance_matrix(channel)
        else:
            print('Calculating covariance in {0} wavelength channels...'.format(nc))
            channels = numpy.linspace(0, rss.nwave-1, num=nc, dtype=int)
            C = rss.covariance_cube(channels=channels)

    print('Writing data to {0}.'.format(ofile))
    C.write(ofile, overwrite=True)            # Write the data


class CalculateCovariance(scriptbase.ScriptBase):

    @classmethod
    def get_parser(cls, width=None):

        parser = super().get_parser(description='Calculate covariance in a MaNGA datacube',
                                    width=width)

        parser.add_argument('plate', type=int, help='plate ID to process')
        parser.add_argument('ifudesign', type=int, help='IFU design to process')
        parser.add_argument('output_file', type=str, help='Name for output file')

        mode = parser.add_mutually_exclusive_group(required=False)
        mode.add_argument('-n', '--numchannels', type=int,
                        help='Number of channels spread across the wavelength range for which '
                            'to compute the covariance matrix.  A value of 0 forces construction '
                            'of the full covariance cube.  The default is to calculate the '
                            'covariance matrix for a single channel at the central wavelength',
                        default=1)
        mode.add_argument('-w', '--wavelength', type=float,
                        help='Wavelength at which to compute a single covariance matrix; default '
                            'is the central wavelength', default=None)

        parser.add_argument('-d', '--directory_path', type=str, default=None,
                            help='Directory with the DRP produced RSS file; default uses '
                                 'environmental variables to define the default MaNGA DRP '
                                 'redux path')
        return parser

    @staticmethod
    def main(args):

        import os
        import time

        t = time.perf_counter()
        if os.path.isfile(args.output_file):
            print('WARNING: Overwriting existing file {0}!'.format(args.output_file))
        calculate_covariance_cube(args.plate, args.ifudesign, args.output_file, nc=args.numchannels,
                                wave=args.wavelength, directory_path=args.directory_path)
        print('Elapsed time: {0} seconds'.format(time.perf_counter() - t))





import numpy
from scipy import stats

from mangadap.proc import bandpassfilter

#-----------------------------------------------------------------------------


def test_passband_integral():
    x = numpy.arange(3)
    y = numpy.ones(2)
    Y = bandpassfilter.passband_integral(x,y,borders=True)
    E = bandpassfilter.passband_integral(x,y,borders=True, quad=True)
    assert numpy.all(numpy.isclose([Y,E], [2.0, numpy.sqrt(2)])), 'Bad flat integral'
    Y = bandpassfilter.passband_integral(x,y,borders=True,passband=[0.5,1.5])
    E = bandpassfilter.passband_integral(x,y,borders=True, quad=True,passband=[0.5,1.5])
    assert numpy.all(numpy.isclose([Y,E], [1.0, 1/numpy.sqrt(2)])), 'Bad passband integral'


def test_passband_width():
    x = numpy.arange(5)
    y = numpy.ma.MaskedArray(numpy.ones(4))
    y[2] = numpy.ma.masked
    assert numpy.all(numpy.isclose(bandpassfilter.passband_integrated_width(x,y,borders=True,
                                   passband=numpy.array([[0,3],[0.5,2.5],[3,3.2]])),
                                   numpy.array([2,1.5,0.2]))), 'Incorrect widths'


def test_passband_mean():
    x = numpy.arange(5)
    y = numpy.ma.MaskedArray(numpy.ones(4))
    y[2] = numpy.ma.masked
    assert numpy.all(numpy.isclose(bandpassfilter.passband_integrated_mean(x,y,borders=True,
                                   passband=numpy.array([[0,3],[0.5,2.5],[3,3.2]]))[0],
                                   numpy.ones(3))), 'Incorrect means'


def test_gaussian_integral():
    x = numpy.linspace(-6,6,100)
    g = stats.norm.pdf(x)
    passbands = numpy.array([[-6,6], [-5,5], [-4,4], [-3,3], [-2,2], [-1,1]])
    assert numpy.all(numpy.isclose(bandpassfilter.passband_integral(x,g,passband=passbands),
                            [1., 0.99999942, 0.99993731, 0.99727786, 0.95423538, 0.68231932])), \
            'Incorrect Gaussian integrals'


def test_passband_weighted_stats():
    x = numpy.linspace(-15,15,500)
    g = stats.norm.pdf(x, loc=-9, scale=0.3) + stats.norm.pdf(x, loc=-3, scale=0.6) \
                + stats.norm.pdf(x, loc=8, scale=1.2)
    p = numpy.array([[-11,-7], [-7,1], [1,15]])
    assert numpy.all(numpy.isclose(bandpassfilter.passband_weighted_mean(x,g,x,passband=p)[0],
                                   [-9, -3, 8])), 'Incorrect weighted mean'
    assert numpy.all(numpy.isclose(bandpassfilter.passband_weighted_sdev(x,g,x,passband=p)[0],
                                   [0.3, 0.6, 1.2])), 'Incorrect weighted standard deviation'


#def test_passband_weighted_stats_errors():
#    # Errors are not really robust enough for a useful test.
#    x = numpy.linspace(-15,15,500)
#    g = stats.norm.pdf(x, loc=-9, scale=0.3) + stats.norm.pdf(x, loc=-3, scale=0.6) \
#                + stats.norm.pdf(x, loc=8, scale=1.2)
#    p = numpy.array([[-11,-7], [-7,1], [1,15]])
#
#    nsim = 1000
#    mom0 = numpy.ma.zeros((nsim,3), dtype=float)
#    mom0e = numpy.ma.zeros((nsim,3), dtype=float)
#    mean = numpy.ma.zeros((nsim,3), dtype=float)
#    meane = numpy.ma.zeros((nsim,3), dtype=float)
#    sdev = numpy.ma.zeros((nsim,3), dtype=float)
#    sdeve = numpy.ma.zeros((nsim,3), dtype=float)
#    err = 0.01
#    e = err*numpy.ones_like(x)
#    for i in range(nsim):
#        n = stats.norm.rvs(scale=0.05, size=x.size)
#        mom0[i,:] = bandpassfilter.passband_integral(x,g+n,passband=p)
#        mom0e[i,:] = bandpassfilter.passband_integral(x,g+n,passband=p,quad=True)
#        mean[i,:], meane[i,:] = bandpassfilter.passband_weighted_mean(x,g+n,x,yerr=e,passband=p)
#        sdev[i,:], sdeve[i,:] = bandpassfilter.passband_weighted_sdev(x,g+n,x,yerr=e,passband=p)
#
#    print('Mom0:')
#    print(numpy.mean(mom0, axis=0))
#    print(numpy.std(mom0, axis=0))
#    print(numpy.mean(mom0e, axis=0))
#    print(numpy.std(mom0e, axis=0))
#
#    print('Mean:')
#    print(numpy.mean(mean, axis=0))
#    print(numpy.std(mean, axis=0))
#    print(numpy.mean(meane, axis=0))
#    print(numpy.std(meane, axis=0))
#
#    print('SDEV:')
#    print(numpy.mean(sdev, axis=0))
#    print(numpy.std(sdev, axis=0))
#    print(numpy.mean(sdeve, axis=0))
#    print(numpy.std(sdeve, axis=0))
#
#    from matplotlib import pyplot
#
#    pyplot.hist(mean[:,0].compressed(), range=[-15,15], bins=500)
#    pyplot.hist(mean[:,1].compressed(), range=[-15,15], bins=500)
#    pyplot.hist(mean[:,2].compressed(), range=[-15,15], bins=500)
#    pyplot.show()
#
#    pyplot.hist(sdev[:,0].compressed(), range=[-15,15], bins=500)
#    pyplot.hist(sdev[:,1].compressed(), range=[-15,15], bins=500)
#    pyplot.hist(sdev[:,2].compressed(), range=[-15,15], bins=500)
#    pyplot.show()
    

#if __name__ == '__main__':
#    test_passband_weighted_stats_errors()
#    test_moments()
    



import time

from IPython import embed

import numpy

from matplotlib import pyplot, rc, colors, colorbar, ticker, cm

from astropy.io import fits

from mangadap import dapfits
from mangadap.datacube import DataCube
from mangadap.config.analysisplan import AnalysisPlan
from mangadap.util.fileio import channel_dictionary
from mangadap.proc.util import growth_lim
from mangadap.util.mapping import map_extent, map_beam_patch
from mangadap.util.pkg import load_object

from mangadap.scripts import scriptbase


def init_ax(fig, pos):
    ax = fig.add_axes(pos, facecolor='0.9')
    ax.minorticks_on()
    ax.tick_params(which='major', length=6, direction='in')
    ax.tick_params(which='minor', length=3, direction='in')
    ax.grid(True, which='major', color='0.8', zorder=0, linestyle='-')
#    ax.grid(True, which='minor', color='0.7', zorder=0, linestyle=':')
    return ax


def masked_imshow(fig, ax, cax, data, extent=None, norm=None, vmin=None, vmax=None, cmap='viridis',
                  zorder=0, cbformat=None, subs=None, orientation='vertical', contour_data=None,
                  levels=None, cblocator=None, cbticks=None):
    if numpy.sum(data.mask) != numpy.prod(data.shape):
        if norm is None:
            img = ax.imshow(data, origin='lower', interpolation='nearest', extent=extent,
                            vmin=vmin, vmax=vmax, cmap=cmap, zorder=zorder)
        else:
            img = ax.imshow(data, origin='lower', interpolation='nearest', extent=extent,
                            norm=norm, cmap=cmap, zorder=zorder)

        cb = fig.colorbar(img, cax=cax, orientation=orientation) if cbformat is None else \
                    fig.colorbar(img, cax=cax, format=cbformat, orientation=orientation)
        if subs is not None:
            cb.locator = ticker.LogLocator(base=10, subs=(1.,2.,4.,))
            cb.update_ticks()
        elif cblocator is not None:
            cb.locator = cblocator
            cb.update_ticks()
        elif cbticks is not None:
            cb.set_ticks(cbticks)
            cb.update_ticks()
    else:
        _norm = colors.Normalize(vmin=vmin, vmax=vmax) if norm is None else norm
        cb = colorbar.ColorbarBase(cax, cmap=cm.get_cmap(cmap), norm=_norm)

    if contour_data is not None and levels is not None \
            and numpy.sum(contour_data.mask) != numpy.prod(contour_data.shape):
        cnt = ax.contour(contour_data, origin='lower', extent=extent, colors='0.5',
                         levels=levels, linewidths=1.5, zorder=5)


# Test images:
#  - SPX_MFLUX, SPX_SNR, BINID, BIN_SNR
#  - STELLAR_CONT_FRESID, STELLAR_CONT_CHI2, STELLAR_VEL, STELLAR_SIGMA
#  - EMLINE_SFLUX, EMLINE_GFLUX, EMLINE_GVEL, EMLINE_GSIGMA - all H-alpha
#  - EMLINE_SFLUX (H-beta), EMLINE_GFLUX (H-beta), D4000, Dn4000
def spotcheck_images(cube, plan, method_dir, ref_dir, qa_dir, fwhm=None):
    """
    Construct a QA plot for the PPXFFit results.

    Args:
        cube (:class:`~mangadap.datacube.datacube.DataCube`):
            Analyzed data cube
        plan (:obj:`dict`, :class:`~mangadap.par.parset.ParSet`):
            Object with analysis plan keywords.
        method_dir (`Path`_):
            Top-level "method" directory with all the DAP files.
        ref_dir (`Path`_):
            Directory with all the DAP reference files.
        qa_dir (`Path`_):
            Directory for the QA figures
        fwhm (:obj:`float`, optional):
            The FWHM in arcsec of the beam.  If None, no beam included in the
            plot.
    """
    # Check that the paths exists
    if not method_dir.exists():
        raise NotADirectoryError(f'{method_dir} does not exist; run the DAP first!')
    if not ref_dir.exists():
        raise NotADirectoryError(f'{ref_dir} does not exist; run the DAP first!')
    if not qa_dir.exists():
        qa_dir.mkdir(parents=True)

    # Get the name of the output file
    ofile = qa_dir / f'{cube.output_root}-MAPS-{plan["key"]}-spotcheck.png'

    bm = dapfits.DAPMapsBitMask()

    _, directory, file = dapfits.default_paths(cube, method=plan['key'], output_path=method_dir)
    ifile = directory / file
    if not ifile.exists():
        raise FileNotFoundError(f'No file: {ifile}')

    print(f'Reading: {ifile}')
    hdu = fits.open(str(ifile))

    # Check everything is finite
    num_ext = len(hdu)
    for i in range(num_ext):
        if hdu[i].data is None:
            continue
        if not numpy.all(numpy.isfinite(hdu[i].data)):
            raise ValueError(f'HDU {hdu[i].name} contains infs or NaNs!')

    # Build the column dictionaries
    emline = channel_dictionary(hdu, 'EMLINE_GFLUX')
    specindex = channel_dictionary(hdu, 'SPECINDEX')

    extent = map_extent(hdu, 'SPX_MFLUX')

    # Get the data to plot
    spxflx = numpy.ma.MaskedArray(hdu['SPX_MFLUX'].data.copy())
    spxsnr = numpy.ma.MaskedArray(hdu['SPX_SNR'].data.copy())
    binid = numpy.ma.MaskedArray(hdu['BINID'].data.copy(), mask=hdu['BINID'].data<0)
    binsnr = numpy.ma.MaskedArray(hdu['BIN_SNR'].data.copy(), mask=hdu['BIN_MFLUX_MASK'].data>0)

    # 68% growth of the absolute value of the fractional residuals
    scfres = numpy.ma.MaskedArray(hdu['STELLAR_FOM'].data.copy()[3,:,:],
                                  mask=numpy.invert(hdu['STELLAR_FOM'].data[3,:,:] > 0))
    # Reduced chi-square
    scrchi = numpy.ma.MaskedArray(hdu['STELLAR_FOM'].data.copy()[2,:,:],
                                  mask=numpy.invert(hdu['STELLAR_FOM'].data[2,:,:] > 0))

    strvel = numpy.ma.MaskedArray(hdu['STELLAR_VEL'].data.copy(),
                                  mask=bm.flagged(hdu['STELLAR_VEL_MASK'].data, 'DONOTUSE'))
    ustrvel = numpy.ma.MaskedArray(hdu['STELLAR_VEL'].data.copy(),
                                  mask=numpy.invert(bm.flagged(hdu['STELLAR_VEL_MASK'].data,
                                                               'UNRELIABLE')))
    ustrvel[numpy.invert(numpy.ma.getmaskarray(ustrvel))] = 0.0
    strsig = numpy.ma.MaskedArray(hdu['STELLAR_SIGMA'].data.copy(),
                                  mask=bm.flagged(hdu['STELLAR_SIGMA_MASK'].data, 'DONOTUSE'))
    ustrsig = numpy.ma.MaskedArray(hdu['STELLAR_SIGMA'].data.copy(),
                                  mask=numpy.invert(bm.flagged(hdu['STELLAR_SIGMA_MASK'].data,
                                                               'UNRELIABLE')))
    ustrsig[numpy.invert(numpy.ma.getmaskarray(ustrsig))] = 0.0

    hasflx = numpy.ma.MaskedArray(hdu['EMLINE_SFLUX'].data.copy()[emline['Ha-6564'],:,:],
            mask=bm.flagged(hdu['EMLINE_SFLUX_MASK'].data[emline['Ha-6564'],:,:], 'DONOTUSE'))
    uhasflx = numpy.ma.MaskedArray(hdu['EMLINE_SFLUX'].data.copy()[emline['Ha-6564'],:,:],
          mask=numpy.invert(bm.flagged(hdu['EMLINE_SFLUX_MASK'].data[emline['Ha-6564'],:,:],
                                       'UNRELIABLE')))
    uhasflx[numpy.invert(numpy.ma.getmaskarray(uhasflx))] = 0.0
    hagflx = numpy.ma.MaskedArray(hdu['EMLINE_GFLUX'].data.copy()[emline['Ha-6564'],:,:],
          mask=bm.flagged(hdu['EMLINE_GFLUX_MASK'].data[emline['Ha-6564'],:,:], 'DONOTUSE'))
    uhagflx = numpy.ma.MaskedArray(hdu['EMLINE_GFLUX'].data.copy()[emline['Ha-6564'],:,:],
          mask=numpy.invert(bm.flagged(hdu['EMLINE_GFLUX_MASK'].data[emline['Ha-6564'],:,:],
                                       'UNRELIABLE')))
    uhagflx[numpy.invert(numpy.ma.getmaskarray(uhagflx))] = 0.0
    hagvel = numpy.ma.MaskedArray(hdu['EMLINE_GVEL'].data.copy()[emline['Ha-6564'],:,:],
          mask=bm.flagged(hdu['EMLINE_GVEL_MASK'].data[emline['Ha-6564'],:,:], 'DONOTUSE'))
    uhagvel = numpy.ma.MaskedArray(hdu['EMLINE_GVEL'].data.copy()[emline['Ha-6564'],:,:],
          mask=numpy.invert(bm.flagged(hdu['EMLINE_GVEL_MASK'].data[emline['Ha-6564'],:,:],
                                       'UNRELIABLE')))
    uhagvel[numpy.invert(numpy.ma.getmaskarray(uhagvel))] = 0.0
    hagsig = numpy.ma.MaskedArray(hdu['EMLINE_GSIGMA'].data.copy()[emline['Ha-6564'],:,:],
        mask=bm.flagged(hdu['EMLINE_GSIGMA_MASK'].data[emline['Ha-6564'],:,:], 'DONOTUSE'))
    uhagsig = numpy.ma.MaskedArray(hdu['EMLINE_GSIGMA'].data.copy()[emline['Ha-6564'],:,:],
        mask=numpy.invert(bm.flagged(hdu['EMLINE_GSIGMA_MASK'].data[emline['Ha-6564'],:,:],
                                     'UNRELIABLE')))
    uhagsig[numpy.invert(numpy.ma.getmaskarray(uhagsig))] = 0.0

    hbsflx = numpy.ma.MaskedArray(hdu['EMLINE_SFLUX'].data.copy()[emline['Hb-4862'],:,:],
          mask=bm.flagged(hdu['EMLINE_SFLUX_MASK'].data[emline['Hb-4862'],:,:], 'DONOTUSE'))
    uhbsflx = numpy.ma.MaskedArray(hdu['EMLINE_SFLUX'].data.copy()[emline['Hb-4862'],:,:],
          mask=numpy.invert(bm.flagged(hdu['EMLINE_SFLUX_MASK'].data[emline['Hb-4862'],:,:],
                                       'UNRELIABLE')))
    uhbsflx[numpy.invert(numpy.ma.getmaskarray(uhbsflx))] = 0.0
    hbgflx = numpy.ma.MaskedArray(hdu['EMLINE_GFLUX'].data.copy()[emline['Hb-4862'],:,:],
          mask=bm.flagged(hdu['EMLINE_GFLUX_MASK'].data[emline['Hb-4862'],:,:], 'DONOTUSE'))
    uhbgflx = numpy.ma.MaskedArray(hdu['EMLINE_GFLUX'].data.copy()[emline['Hb-4862'],:,:],
          mask=numpy.invert(bm.flagged(hdu['EMLINE_GFLUX_MASK'].data[emline['Hb-4862'],:,:],
                                       'UNRELIABLE')))
    uhbgflx[numpy.invert(numpy.ma.getmaskarray(uhbgflx))] = 0.0

    specindex_available = True
    try:
        if hdu['SPECINDEX'].data is None:
            pass
    except:
        warnings.warn('No spectral indices in maps file.')
        specindex_available = False
    if specindex_available:
        d4000 = numpy.ma.MaskedArray(hdu['SPECINDEX'].data.copy()[specindex['D4000'],:,:],
                                mask=bm.flagged(hdu['SPECINDEX_MASK'].data[specindex['D4000'],:,:],
                                                'DONOTUSE'))
        ud4000 = numpy.ma.MaskedArray(hdu['SPECINDEX'].data.copy()[specindex['D4000'],:,:],
                                mask=numpy.invert(bm.flagged(hdu['SPECINDEX_MASK'].data[
                                                      specindex['D4000'],:,:], 'UNRELIABLE')))
        ud4000[numpy.invert(numpy.ma.getmaskarray(ud4000))] = 0.0
        dn4000 = numpy.ma.MaskedArray(hdu['SPECINDEX'].data.copy()[specindex['Dn4000'],:,:],
                                mask=bm.flagged(hdu['SPECINDEX_MASK'].data[specindex['Dn4000'],:,:],
                                                'DONOTUSE'))
        udn4000 = numpy.ma.MaskedArray(hdu['SPECINDEX'].data.copy()[specindex['Dn4000'],:,:],
                                mask=numpy.invert(bm.flagged(
                                        hdu['SPECINDEX_MASK'].data[specindex['Dn4000'],:,:],
                                        'UNRELIABLE')))
        udn4000[numpy.invert(numpy.ma.getmaskarray(udn4000))] = 0.0
    else:
        output_shape = spxflx.shape
        d4000 = numpy.ma.masked_all(output_shape)
        ud4000 = numpy.ma.masked_all(output_shape)
        dn4000 = numpy.ma.masked_all(output_shape)
        udn4000 = numpy.ma.masked_all(output_shape)

    # Get the limits to apply
    x_order = 1 if extent[0] < extent[1] else -1
    Dx = x_order*(extent[1] - extent[0])
    y_order = 1 if extent[2] < extent[3] else -1
    Dy = y_order*(extent[3] - extent[2])
    # Force the image panels to be square; assumes extent has the same units in
    # both dimensions.
    D = max(Dx, Dy)

    x_lim = (extent[0]+extent[1])/2 + D * x_order * numpy.array([-1,1])/2
    y_lim = (extent[2]+extent[3])/2 + D * y_order * numpy.array([-1,1])/2

    flux_lim = numpy.power(10, growth_lim(numpy.ma.log10(spxflx), 0.90, fac=1.05))
    t = numpy.ma.append(numpy.ma.log10(spxsnr), numpy.ma.log10(binsnr))
    snr_lim = numpy.power(10, growth_lim(t, 0.90, fac=1.05))
    bin_lim = [ numpy.ma.amin(binid), numpy.ma.amax(binid) ]

    res_lim = numpy.power(10, growth_lim(numpy.ma.log10(scfres), 0.90, fac=1.05))
    chi_lim = growth_lim(scrchi, 0.90, fac=1.05)
    t = numpy.ma.append(strvel, hagvel)
    vel_lim = growth_lim(t, 0.85, fac=1.10, midpoint=0.0)
    if vel_lim[0] < -350:
        vel_lim[0] = -350
    if vel_lim[1] > 350:
        vel_lim[1] = 350
    t = numpy.ma.append(numpy.ma.log10(strsig), numpy.ma.log10(hagsig))
    sig_lim = numpy.power(10, growth_lim(t, 0.90, fac=1.05))

    t = numpy.ma.append( numpy.ma.append(numpy.ma.log10(hasflx), numpy.ma.log10(hagflx)),
                         numpy.ma.append(numpy.ma.log10(hbsflx), numpy.ma.log10(hbgflx)) )
#    t = numpy.ma.log10(t)
    hflx_lim = numpy.power(10, growth_lim(t, 0.95, fac=1.10))
#    hflx_lim = [ -0.3, 2.2 ]
    t = numpy.ma.append(d4000, dn4000)
#    d4000_lim = growth_lim(t, 0.90, 1.05)
    d4000_lim = [ 1.0, 2.5 ]

    font = { 'size' : 6 }
    rc('font', **font)

    w,h = pyplot.figaspect(1)
    fig = pyplot.figure(figsize=(1.5*w,1.5*h))

    dx = 0.22
    left = 0.05
    top=0.94
    dw = 0.005

    snr_levels = [0.5, 1.0, 1.5, 2.0 ]

    ax = init_ax(fig, [left, top-dx, dx, dx])
    ax.set_xlim(x_lim)
    ax.set_ylim(y_lim)
    ax.xaxis.set_major_formatter(ticker.NullFormatter())
    cax = fig.add_axes([left+dx*0.65, top-0.07*dx, 0.27*dx, 0.01])
    cbticks = [flux_lim[0]*1.1, flux_lim[1]/1.1]
    masked_imshow(fig, ax, cax, spxflx, extent=extent,
                  norm=colors.LogNorm(vmin=flux_lim[0], vmax=flux_lim[1]), cmap='YlGnBu_r',
                  zorder=4, orientation='horizontal', contour_data=numpy.ma.log10(spxsnr),
                  levels=snr_levels, cbticks=cbticks, cbformat=ticker.ScalarFormatter())
    if fwhm is not None:
        ax.add_patch(map_beam_patch(extent, ax, fwhm=fwhm, facecolor='0.7', edgecolor='k', zorder=4))
    ax.text(0.99, 0.05, r'g-band S (spx)', horizontalalignment='right',
            verticalalignment='center', transform=ax.transAxes)

    ax = init_ax(fig, [left+dx+dw, top-dx, dx, dx])
    ax.set_xlim(x_lim)
    ax.set_ylim(y_lim)
    ax.xaxis.set_major_formatter(ticker.NullFormatter())
    ax.yaxis.set_major_formatter(ticker.NullFormatter())
    cax = fig.add_axes([left+dx+dw+dx*0.65, top-0.07*dx, 0.27*dx, 0.01])
    cbticks = [snr_lim[0]*1.1, snr_lim[1]/1.1]
    masked_imshow(fig, ax, cax, spxsnr, extent=extent,
                  norm=colors.LogNorm(vmin=snr_lim[0], vmax=snr_lim[1]), cmap='YlGnBu_r',
                  zorder=4, orientation='horizontal', contour_data=numpy.ma.log10(spxsnr),
                  levels=snr_levels, cbticks=cbticks, cbformat=ticker.ScalarFormatter()) #'%.0f')
    if fwhm is not None:
        ax.add_patch(map_beam_patch(extent, ax, fwhm=fwhm, facecolor='0.7', edgecolor='k', zorder=4))
    ax.text(0.99, 0.05, r'g-band S/N (spx)', horizontalalignment='right',
            verticalalignment='center', transform=ax.transAxes)
   
#    ax.text(1.0, 1.1, r'{0}-{1}; {2}'.format(plate,ifudesign, hdu['PRIMARY'].header['MANGAID']),
#            horizontalalignment='center', verticalalignment='center', transform=ax.transAxes,
#            fontsize=20)
    ax.text(1.0, 1.1, f'{cube.name}; {plan["key"]}', ha='center', va='center',
            transform=ax.transAxes, fontsize=20)

    ax = init_ax(fig, [left+2*dx+2*dw, top-dx, dx, dx])
    ax.set_xlim(x_lim)
    ax.set_ylim(y_lim)
    ax.xaxis.set_major_formatter(ticker.NullFormatter())
    ax.yaxis.set_major_formatter(ticker.NullFormatter())
    cax = fig.add_axes([left+2*dx+2*dw+dx*0.65, top-0.07*dx, 0.27*dx, 0.01])
    masked_imshow(fig, ax, cax, binid[0,:,:], extent=extent,
                  vmin=bin_lim[0], vmax=bin_lim[1], cmap='CMRmap',
                  zorder=4, orientation='horizontal', contour_data=numpy.ma.log10(spxsnr),
                  levels=snr_levels, cbformat=ticker.ScalarFormatter(),
                  cbticks=bin_lim) 
    if fwhm is not None:
        ax.add_patch(map_beam_patch(extent, ax, fwhm=fwhm, facecolor='0.7', edgecolor='k', zorder=4))
    ax.text(0.99, 0.05, r'bin ID', horizontalalignment='right',
            verticalalignment='center', transform=ax.transAxes)
 
    ax = init_ax(fig, [left+3*dx+3*dw, top-dx, dx, dx])
    ax.set_xlim(x_lim)
    ax.set_ylim(y_lim)
    ax.xaxis.set_major_formatter(ticker.NullFormatter())
    ax.yaxis.set_major_formatter(ticker.NullFormatter())
    cax = fig.add_axes([left+3*dx+3*dw+dx*0.65, top-0.07*dx, 0.27*dx, 0.01])
    cbticks = [snr_lim[0]*1.1, snr_lim[1]/1.1]
    masked_imshow(fig, ax, cax, binsnr, extent=extent,
                  norm=colors.LogNorm(vmin=snr_lim[0], vmax=snr_lim[1]), cmap='YlGnBu_r',
                  zorder=4, orientation='horizontal', contour_data=numpy.ma.log10(spxsnr),
                  levels=snr_levels, cbticks=cbticks, cbformat=ticker.ScalarFormatter()) #'%.0f')
    if fwhm is not None:
        ax.add_patch(map_beam_patch(extent, ax, fwhm=fwhm, facecolor='0.7', edgecolor='k', zorder=4))
    ax.text(0.99, 0.05, r'g-band S/N (bin)', horizontalalignment='right',
            verticalalignment='center', transform=ax.transAxes)


    ax = init_ax(fig, [left, top-2*dx-dw, dx, dx])
    ax.set_xlim(x_lim)
    ax.set_ylim(y_lim)
    ax.xaxis.set_major_formatter(ticker.NullFormatter())
    cax = fig.add_axes([left+dx*0.65, top-dx-dw-0.07*dx, 0.27*dx, 0.01])
    cbticks = [res_lim[0]*1.1, res_lim[1]/1.1]
    masked_imshow(fig, ax, cax, scfres, extent=extent,
                  norm=colors.LogNorm(vmin=res_lim[0], vmax=res_lim[1]), cmap='inferno',
                  zorder=4, orientation='horizontal', contour_data=numpy.ma.log10(spxsnr),
                  levels=snr_levels, cbticks=cbticks, cbformat=ticker.ScalarFormatter()) #'%.0f')
    if fwhm is not None:
        ax.add_patch(map_beam_patch(extent, ax, fwhm=fwhm, facecolor='0.7', edgecolor='k', zorder=4))
    ax.text(0.99, 0.05, r'pPXF Frac. Resid.', horizontalalignment='right',
            verticalalignment='center', transform=ax.transAxes)

    ax = init_ax(fig, [left+dx+dw,  top-2*dx-dw, dx, dx])
    ax.set_xlim(x_lim)
    ax.set_ylim(y_lim)
    ax.xaxis.set_major_formatter(ticker.NullFormatter())
    ax.yaxis.set_major_formatter(ticker.NullFormatter())
    cax = fig.add_axes([left+dx+dw+dx*0.65, top-dx-dw-0.07*dx, 0.27*dx, 0.01])
    cbticks=[numpy.round(chi_lim[0]+0.01, decimals=2), numpy.round(chi_lim[1]-0.01, decimals=2)]
    masked_imshow(fig, ax, cax, scrchi, extent=extent,
                  vmin=chi_lim[0], vmax=chi_lim[1], cmap='inferno',
                  zorder=4, orientation='horizontal', contour_data=numpy.ma.log10(spxsnr),
                  levels=snr_levels, cbticks=cbticks, cbformat=ticker.ScalarFormatter()) #'%.0f')
    if fwhm is not None:
        ax.add_patch(map_beam_patch(extent, ax, fwhm=fwhm, facecolor='0.7', edgecolor='k', zorder=4))
    ax.text(0.99, 0.05, r'${\rm pPXF}\ \chi^2_\nu$', horizontalalignment='right',
            verticalalignment='center', transform=ax.transAxes)
    
    ax = init_ax(fig, [left+2*dx+2*dw,  top-2*dx-dw, dx, dx])
    ax.set_xlim(x_lim)
    ax.set_ylim(y_lim)
    ax.xaxis.set_major_formatter(ticker.NullFormatter())
    ax.yaxis.set_major_formatter(ticker.NullFormatter())
    cax = fig.add_axes([left+2*dx+2*dw+dx*0.65, top-dx-dw-0.07*dx, 0.27*dx, 0.01])
    cbticks = [vel_lim[0]/1.1, 0.0, vel_lim[1]/1.1]
    masked_imshow(fig, ax, cax, strvel, extent=extent,
                  vmin=vel_lim[0], vmax=vel_lim[1], cmap='RdBu_r',
                  zorder=4, orientation='horizontal', contour_data=numpy.ma.log10(spxsnr),
                  levels=snr_levels, cbticks=cbticks, cbformat=ticker.ScalarFormatter()) #'%.0f')
    if fwhm is not None:
        ax.add_patch(map_beam_patch(extent, ax, fwhm=fwhm, facecolor='0.7', edgecolor='k', zorder=4))
    ax.text(0.99, 0.05, r'$V_\ast$', horizontalalignment='right',
            verticalalignment='center', transform=ax.transAxes)
    
    ax = init_ax(fig, [left+3*dx+3*dw,  top-2*dx-dw, dx, dx])
    ax.set_xlim(x_lim)
    ax.set_ylim(y_lim)
    ax.xaxis.set_major_formatter(ticker.NullFormatter())
    ax.yaxis.set_major_formatter(ticker.NullFormatter())
    cax = fig.add_axes([left+3*dx+3*dw+dx*0.65, top-dx-dw-0.07*dx, 0.27*dx, 0.01])
    cbticks = [sig_lim[0]*1.1, numpy.power(10, numpy.mean(numpy.log10(sig_lim))), sig_lim[1]/1.1]
    masked_imshow(fig, ax, cax, strsig, extent=extent,
                  norm=colors.LogNorm(vmin=sig_lim[0], vmax=sig_lim[1]), cmap='viridis',
                  zorder=4, orientation='horizontal', contour_data=numpy.ma.log10(spxsnr),
                  levels=snr_levels, cbticks=cbticks, cbformat=ticker.ScalarFormatter()) #'%.0f')
    if fwhm is not None:
        ax.add_patch(map_beam_patch(extent, ax, fwhm=fwhm, facecolor='0.7', edgecolor='k', zorder=4))
    ax.text(0.99, 0.05, r'$\sigma_\ast$', horizontalalignment='right',
            verticalalignment='center', transform=ax.transAxes)



    ax = init_ax(fig, [left, top-3*dx-2*dw, dx, dx])
    ax.set_xlim(x_lim)
    ax.set_ylim(y_lim)
    ax.xaxis.set_major_formatter(ticker.NullFormatter())
    cax = fig.add_axes([left+dx*0.65, top-2*dx-2*dw-0.07*dx, 0.27*dx, 0.01])
    cbticks = [hflx_lim[0]*1.1, hflx_lim[1]/1.1]
    masked_imshow(fig, ax, cax, hasflx, extent=extent,
                  norm=colors.LogNorm(vmin=hflx_lim[0], vmax=hflx_lim[1]), cmap='inferno',
                  zorder=4, orientation='horizontal', contour_data=numpy.ma.log10(spxsnr),
                  levels=snr_levels, cbticks=cbticks, cbformat=ticker.ScalarFormatter()) #'%.0f')
    if fwhm is not None:
        ax.add_patch(map_beam_patch(extent, ax, fwhm=fwhm, facecolor='0.7', edgecolor='k', zorder=4))
    ax.text(0.99, 0.05, r'${\rm H}\alpha\ {\rm flux}$ (sum)', horizontalalignment='right',
            verticalalignment='center', transform=ax.transAxes)

    ax = init_ax(fig, [left+dx+dw,  top-3*dx-2*dw, dx, dx])
    ax.set_xlim(x_lim)
    ax.set_ylim(y_lim)
    ax.xaxis.set_major_formatter(ticker.NullFormatter())
    ax.yaxis.set_major_formatter(ticker.NullFormatter())
    cax = fig.add_axes([left+dx+dw+dx*0.65, top-2*dx-2*dw-0.07*dx, 0.27*dx, 0.01])
    cbticks = [hflx_lim[0]*1.1, hflx_lim[1]/1.1]
    masked_imshow(fig, ax, cax, hagflx, extent=extent,
                  norm=colors.LogNorm(vmin=hflx_lim[0], vmax=hflx_lim[1]), cmap='inferno',
                  zorder=4, orientation='horizontal', contour_data=numpy.ma.log10(spxsnr),
                  levels=snr_levels, cbticks=cbticks, cbformat=ticker.ScalarFormatter()) #'%.0f')
    if fwhm is not None:
        ax.add_patch(map_beam_patch(extent, ax, fwhm=fwhm, facecolor='0.7', edgecolor='k', zorder=4))
    ax.text(0.99, 0.05, r'${\rm H}\alpha\ {\rm flux}$ (Gauss)', horizontalalignment='right',
            verticalalignment='center', transform=ax.transAxes)
    
    ax = init_ax(fig, [left+2*dx+2*dw,  top-3*dx-2*dw, dx, dx])
    ax.set_xlim(x_lim)
    ax.set_ylim(y_lim)
    ax.xaxis.set_major_formatter(ticker.NullFormatter())
    ax.yaxis.set_major_formatter(ticker.NullFormatter())
    cax = fig.add_axes([left+2*dx+2*dw+dx*0.65, top-2*dx-2*dw-0.07*dx, 0.27*dx, 0.01])
    cbticks = [vel_lim[0]/1.1, 0.0, vel_lim[1]/1.1]
    masked_imshow(fig, ax, cax, hagvel, extent=extent,
                  vmin=vel_lim[0], vmax=vel_lim[1], cmap='RdBu_r',
                  zorder=4, orientation='horizontal', contour_data=numpy.ma.log10(spxsnr),
                  levels=snr_levels, cbticks=cbticks, cbformat=ticker.ScalarFormatter()) #'%.0f')
    if fwhm is not None:
        ax.add_patch(map_beam_patch(extent, ax, fwhm=fwhm, facecolor='0.7', edgecolor='k', zorder=4))
    ax.text(0.99, 0.05, r'$V_{{\rm H}\alpha}$', horizontalalignment='right',
            verticalalignment='center', transform=ax.transAxes)
    
    ax = init_ax(fig, [left+3*dx+3*dw,  top-3*dx-2*dw, dx, dx])
    ax.set_xlim(x_lim)
    ax.set_ylim(y_lim)
    ax.xaxis.set_major_formatter(ticker.NullFormatter())
    ax.yaxis.set_major_formatter(ticker.NullFormatter())
    cax = fig.add_axes([left+3*dx+3*dw+dx*0.65, top-2*dx-2*dw-0.07*dx, 0.27*dx, 0.01])
    cbticks = [sig_lim[0]*1.1, numpy.power(10, numpy.mean(numpy.log10(sig_lim))), sig_lim[1]/1.1]
    masked_imshow(fig, ax, cax, hagsig, extent=extent,
                  norm=colors.LogNorm(vmin=sig_lim[0], vmax=sig_lim[1]), cmap='viridis',
                  zorder=4, orientation='horizontal', contour_data=numpy.ma.log10(spxsnr),
                  levels=snr_levels, cbticks=cbticks, cbformat=ticker.ScalarFormatter()) #'%.0f')
    if fwhm is not None:
        ax.add_patch(map_beam_patch(extent, ax, fwhm=fwhm, facecolor='0.7', edgecolor='k', zorder=4))
    ax.text(0.99, 0.05, r'$\sigma_{{\rm H}\alpha}$', horizontalalignment='right',
            verticalalignment='center', transform=ax.transAxes)



    ax = init_ax(fig, [left, top-4*dx-3*dw, dx, dx])
    ax.set_xlim(x_lim)
    ax.set_ylim(y_lim)
    cax = fig.add_axes([left+dx*0.65, top-3*dx-3*dw-0.07*dx, 0.27*dx, 0.01])
    cbticks = [hflx_lim[0]*1.1, hflx_lim[1]/1.1]
    masked_imshow(fig, ax, cax, hbsflx, extent=extent,
                  norm=colors.LogNorm(vmin=hflx_lim[0], vmax=hflx_lim[1]), cmap='inferno',
                  zorder=4, orientation='horizontal', contour_data=numpy.ma.log10(spxsnr),
                  levels=snr_levels, cbticks=cbticks, cbformat=ticker.ScalarFormatter()) #'%.0f')
    if fwhm is not None:
        ax.add_patch(map_beam_patch(extent, ax, fwhm=fwhm, facecolor='0.7', edgecolor='k', zorder=4))
    ax.text(0.99, 0.05, r'${\rm H}\beta\ {\rm flux}$ (sum)', horizontalalignment='right',
            verticalalignment='center', transform=ax.transAxes)

    ax = init_ax(fig, [left+dx+dw,  top-4*dx-3*dw, dx, dx])
    ax.set_xlim(x_lim)
    ax.set_ylim(y_lim)
    ax.yaxis.set_major_formatter(ticker.NullFormatter())
    cax = fig.add_axes([left+dx+dw+dx*0.65, top-3*dx-3*dw-0.07*dx, 0.27*dx, 0.01])
    cbticks = [hflx_lim[0]*1.1, hflx_lim[1]/1.1]
    masked_imshow(fig, ax, cax, hbgflx, extent=extent,
                  norm=colors.LogNorm(vmin=hflx_lim[0], vmax=hflx_lim[1]), cmap='inferno',
                  zorder=4, orientation='horizontal', contour_data=numpy.ma.log10(spxsnr),
                  levels=snr_levels, cbticks=cbticks, cbformat=ticker.ScalarFormatter()) #'%.0f')
    if fwhm is not None:
        ax.add_patch(map_beam_patch(extent, ax, fwhm=fwhm, facecolor='0.7', edgecolor='k', zorder=4))
    ax.text(0.99, 0.05, r'${\rm H}\beta\ {\rm flux}$ (Gauss)', horizontalalignment='right',
            verticalalignment='center', transform=ax.transAxes)

    ax = init_ax(fig, [left+2*dx+2*dw,  top-4*dx-3*dw, dx, dx])
    ax.set_xlim(x_lim)
    ax.set_ylim(y_lim)
    ax.yaxis.set_major_formatter(ticker.NullFormatter())
    cax = fig.add_axes([left+2*dx+2*dw+dx*0.65, top-3*dx-3*dw-0.07*dx, 0.27*dx, 0.01])
    masked_imshow(fig, ax, cax, d4000, extent=extent,
                  vmin=d4000_lim[0], vmax=d4000_lim[1], cmap='RdBu_r',
                  zorder=4, orientation='horizontal', contour_data=numpy.ma.log10(spxsnr),
                  levels=snr_levels, cbticks=d4000_lim, cbformat=ticker.ScalarFormatter())
    if fwhm is not None:
        ax.add_patch(map_beam_patch(extent, ax, fwhm=fwhm, facecolor='0.7', edgecolor='k', zorder=4))
    ax.text(0.99, 0.05, r'D4000', horizontalalignment='right',
            verticalalignment='center', transform=ax.transAxes)
    
    ax = init_ax(fig, [left+3*dx+3*dw,  top-4*dx-3*dw, dx, dx])
    ax.set_xlim(x_lim)
    ax.set_ylim(y_lim)
    ax.yaxis.set_major_formatter(ticker.NullFormatter())
    cax = fig.add_axes([left+3*dx+3*dw+dx*0.65, top-3*dx-3*dw-0.07*dx, 0.27*dx, 0.01])
    masked_imshow(fig, ax, cax, dn4000, extent=extent,
                  vmin=d4000_lim[0], vmax=d4000_lim[1], cmap='RdBu_r',
                  zorder=4, orientation='horizontal', contour_data=numpy.ma.log10(spxsnr),
                  levels=snr_levels, cbticks=d4000_lim, cbformat=ticker.ScalarFormatter())
    if fwhm is not None:
        ax.add_patch(map_beam_patch(extent, ax, fwhm=fwhm, facecolor='0.7', edgecolor='k', zorder=4))
    ax.text(0.99, 0.05, r'Dn4000', horizontalalignment='right',
            verticalalignment='center', transform=ax.transAxes)

    if ofile is not None:
        fig.canvas.print_figure(ofile, bbox_inches='tight')
        print('Writing: {0}'.format(ofile))
    else:
        pyplot.show()

    fig.clear()
    pyplot.close(fig)
    

class SpotcheckDapMaps(scriptbase.ScriptBase):

    @classmethod
    def name(cls):
        """
        Return the name of the executable.
        """
        return 'spotcheck_dap_maps'

    @classmethod
    def get_parser(cls, width=None):

        parser = super().get_parser(description='Construct a QA plot to spotcheck DAP results',
                                    width=width)

        group = parser.add_mutually_exclusive_group(required=True)
        group.add_argument('-c', '--config', type=str, default=None,
                           help='Configuration file used to instantiate the relevant DataCube '
                                'derived class.')
        group.add_argument('-f', '--cubefile', type=str, default=None,
                           help='Name of the file with the datacube data.  Must be possible to '
                                'instantiate the relevant DataCube derived class directly from '
                                'the file only.')
        parser.add_argument('--cube_module', nargs='*',
                            default='mangadap.datacube.MaNGADataCube',
                            help='The name of the module that contains the DataCube derived '
                                 'class used to read the data.')

        parser.add_argument('-p', '--plan', type=str,
                            help='SDSS parameter file with analysis plan.  If not provided, a '
                                 'default plan is used.')
        parser.add_argument('--plan_module', nargs='*',
                            default='mangadap.config.manga.MaNGAAnalysisPlan',
                            help='The name of the module used to define the analysis plan and '
                                 'the output paths.')

        parser.add_argument('-o', '--output_path', type=str, default=None,
                            help='Top-level directory for the DAP output files; default path is '
                                 'set by the provided analysis plan object (see plan_module).')

        parser.add_argument('-b', '--beam', type=float, default=None, help='Beam FWHM for plot.')

        parser.add_argument('--normal_backend', dest='bgagg', action='store_false', default=True)

        return parser

    @staticmethod
    def main(args):

        t = time.perf_counter()

        if args.bgagg:
            pyplot.switch_backend('agg')

        # Instantiate the DataCube
        #   - Import the module used to read the datacube
        if isinstance(args.cube_module, list) and len(args.cube_module) > 2:
            raise ValueError('Provided cube module must be one or two strings.')
        if isinstance(args.cube_module, str) or len(args.cube_module) == 1:
            UserDataCube = load_object(args.cube_module if isinstance(args.cube_module, str) 
                                       else args.cube_module[0])
        else:
            UserDataCube = load_object(args.cube_module[0], obj=args.cube_module[1])
        #   - Check that the class is derived from DataCube
        if not issubclass(UserDataCube, DataCube):
            raise TypeError('Defined cube object must subclass from mangadap.datacube.DataCube.')

        #   - Import the module used to set the analysis plan
        if isinstance(args.plan_module, list) and len(args.plan_module) > 2:
            raise ValueError('Provided plan module must be one or two strings.')
        if isinstance(args.plan_module, str) or len(args.plan_module) == 1:
            UserPlan = load_object(args.plan_module if isinstance(args.plan_module, str) 
                                       else args.plan_module[0])
        else:
            UserPlan = load_object(args.plan_module[0], obj=args.plan_module[1])
        #   - Check that the class is derived from AnalysisPlan
        if not issubclass(UserPlan, AnalysisPlan):
            raise TypeError('Defined plan object must subclass from '
                            'mangadap.config.analysisplan.AnalysisPlan')

        #   - Instantiate using either the datacube file directly or a
        #     configuration file
        cube = UserDataCube(args.cubefile) if args.config is None \
                    else UserDataCube.from_config(args.config)

        # Read the analysis plan
        plan = UserPlan.default(cube=cube, analysis_path=args.output_path) if args.plan is None \
                    else UserPlan.from_toml(args.plan, cube=cube, analysis_path=args.output_path)

        # Construct the plot for each analysis plan
        for i, key in enumerate(plan.keys()):
            spotcheck_images(cube, plan[key], plan.method_path(plan_index=i),
                             plan.method_path(plan_index=i, ref=True),
                             plan.method_path(plan_index=i, qa=True), fwhm=args.beam)

        print('Elapsed time: {0} seconds'.format(time.perf_counter() - t))






"""
Module with generic functions used when modeling data.

----

.. include license and copyright
.. include:: ../include/copy.rst

----

.. include common links, assuming primary doc root is up one directory
.. include:: ../include/links.rst
"""

import numpy
from .filter import BoxcarFilter

def reject_residuals_1d(resid, lo=3.0, hi=3.0, boxcar=None):
    r"""
    Reject fit residuals.

    Args:
        resid (`numpy.ndarray`_, numpy.ma.MaskedArray`_):
            Weighted fit residuals. If input as a `numpy.ndarray`_,
            all data is included in the calculation of the local or
            global standard deviation. To exclude values from this
            calculation, the input *must* be a masked array. Shape
            must be either :math:`(N_x,)` or :math:`(N_{\rm
            vec},N_x})`. If 2D, the rejection is performed separately
            for each *row*.
        lo (:obj:`float`, optional):
            Sigma rejection for low values.
        hi (:obj:`float`, optional):
            Sigma rejection for high values.
        boxcar (:obj:`int`, optional):
            Size of a boxcar to use if using local sigma
            rejection. If None, rejection is based on the global
            standard deviation.

    Returns:
        `numpy.ndarray`_: Boolean array flagging values to reject.
    """
    if boxcar is None:
        mean = numpy.ma.mean(resid, axis=-1)
        std = numpy.ma.std(resid, axis=-1)
        # NOTE: Transposes avoid the need for numpy.newaxis usage.
        return ((numpy.ma.asarray(resid).data.T < mean - lo * std) \
                | (numpy.ma.asarray(resid).data.T > mean + hi * std)).T \
                    & numpy.logical_not(numpy.ma.getmaskarray(resid))

    bf = BoxcarFilter(boxcar, lo=lo, hi=hi, niter=1, y=resid, local_sigma=True)
    return numpy.squeeze(bf.output_mask) & numpy.logical_not(numpy.ma.getmaskarray(resid))


# TODO: Should maybe change the behavior of this so that if rng=None,
# the range is set to x[[0,-1]].
def scaled_coordinates(x, rng=None):
    r"""
    Scale the provided coordinates.

    The coordinates are scaled such that over the provided range the
    coordinates are rescaled to the range :math:`[-1, 1]`.

    .. warning::

        If ``rng`` is None, the returned object is identical to ``x``
        (i.e., it's not a copy of it).

    Args:
        x (`numpy.ndarray`_):
            Coordinate array
        rng (array-like, optional):
            The range over which to rescale to [-1,1]. If None, the
            input coordinates are simply returned.

    Returns:
        `numpy.ndarray`_: The rescaled coordinates.

    """
    return x if rng is None else 2 * (x - rng[0]) / (rng[1] - rng[0]) - 1 


class FitResiduals:
    def __init__(self, y, model, gpm=None):
        self.y = y
        self.model = model
        self.gpm = gpm
    def __call__(self, a):
        resid = (self.y-self.model(a))
        return resid if self.gpm is None else resid[self.gpm]


class FitChiSquare:
    def __init__(self, y, e, model, gpm=None):
        self.y = y
        self.e = e
        self.model = model
        _gpm = self.e > 0
        self.gpm = _gpm if gpm is None else gpm & _gpm
    def __call__(self, a):
        return (self.y[self.gpm]-self.model(a)[self.gpm])/self.e[self.gpm]


